"""This submodule provides utilities for pytorch model manipulation."""
import sys
from collections import OrderedDict
from copy import deepcopy
from dataclasses import dataclass, field
from functools import partial
from typing import Any, Iterable

import numpy as np
import torch
from torch import nn
from torch.nn import Module
from torch.nn import functional as F
from torch.optim import Optimizer, lr_scheduler

sys.path.append(".")
sys.path.append("..")


__all__ = [
    "AllLayerOutputModel",
    "ModOpt",
    "MMMixin",
    "LinesLinear",
    "LinesConv2d",
    "LinesLSTM",
    "LinesEmbedding",
    "LinesBN2d",
    "aggregate_models",
    "batch_norm_to_group_norm",
    "check_model_fit_mem",
    "diff_model",
    "flatten_parameters",
    "get_activation_size",
    "get_global_model_dict",
    "get_local_model_dict",
    "get_output_shape",
    "get_trainable_keys",
    "merge_models",
    "mix_networks",
    "set_lambda_model",
    "safe_load_state_dict",
    "state_dict_zero_like"
]

# ("num_batches_tracked", "running_mean", "running_var")
STATE_DICT_KEYS_TO_IGNORE = tuple()


@dataclass
class ModOpt:
    """Dataclass to store the model, its associated optimizer and scheduler.

    Attributes:
        model (Module): The model.
        optimizer (Optimizer): The optimizer.
        scheduler (lr_scheduler._LRScheduler): The scheduler.
    """
    model: Module = field(default=None, metadata={"help": "The model"})
    optimizer: Optimizer = field(default=None, metadata={"help": "The optimizer"})
    scheduler: lr_scheduler._LRScheduler = field(
        default=None, metadata={"help": "The scheduler"})
    additional: dict[str, Any] = field(default=None, metadata={"help": "Additional fields"})

    def state_dict(self) -> dict:
        """Get the state dictionary of the model, optimizer and scheduler, plus additional fields.

        Returns:
            dict: The state dictionary.
        """
        return {
            "model": self.model.state_dict() if self.model is not None else None,
            "optimizer": self.optimizer.state_dict() if self.optimizer is not None else None,
            "scheduler": self.scheduler.state_dict() if self.scheduler is not None else None,
            "additional": self.additional
        }

    def load_state_dict(self, state_dict: dict) -> None:
        """Load the state dictionary into the model, optimizer and scheduler, plus
        additional fields if present.

        Args:
            state_dict (dict): The state dictionary.
        """
        if self.model is not None and "model" in state_dict:
            self.model.load_state_dict(state_dict["model"])
        if self.optimizer is not None and "optimizer" in state_dict:
            self.optimizer.load_state_dict(state_dict["optimizer"])
        if self.scheduler is not None and "scheduler" in state_dict:
            self.scheduler.load_state_dict(state_dict["scheduler"])
        if "additional" in state_dict:
            self.additional = state_dict["additional"]


class MMMixin:
    """Mixin class for model interpolation.
    This class provides the necessary methods to interpolate between two models. This mixin class
    must be used as a parent class for the PyTorch modules that need to be interpolated.

    Tip:
        Ideally, when using this mixin to implement a new class ``M``, this should be mixed with
        a class ``C`` that extends :class:`torch.nn.Module` and the interpolation of the parameters
        should happen between the parameters in ``C`` and a new set of parameters defined in ``A``.
        This type of multiple inheritance must have as first parent the class :class:`MMMixin` and
        as second parent a class that extends :class:`torch.nn.Module`.
        For example:

        .. code-block:: python
            :linenos:

            # C is a class that extends torch.nn.Module
            class M(MMMixin, C):
                def __init__(self, *args, **kwargs: dict[str, Any]):
                    super().__init__(*args, **kwargs)
                    self.weight_local = nn.Parameter(torch.zeros_like(self.weight))

        In this case, the default implementation of the method :meth:`get_weight` will work and
        will interpolate between the ``weight`` and the ``weight_local`` attribute of the module.

    Attributes:
        lam (float): The interpolation constant.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.lam: float = None

    def set_lambda(self, lam) -> None:
        """Set the interpolation constant.

        Args:
            lam (float): The interpolation constant.
        """
        self.lam = lam

    def get_lambda(self) -> float:
        """Get the interpolation constant.

        Returns:
            float: The interpolation constant.
        """
        return self.lam

    # @abstractmethod
    def get_weight(self) -> torch.Tensor:
        """Get the interpolated weights of the layer or module according to the interpolation
        constant :attr:`lam`. The default implementation assumes that the layer or module has a
        ``weight`` attribute and a ``weight_local`` attribute that are both tensors of the same
        shape. The interpolated weights are computed as:
        ``w = (1 - self.lam) * self.weight + self.lam * self.weight_local``.

        Returns:
            torch.Tensor: The interpolated weights.
        """
        w = (1 - self.lam) * self.weight + self.lam * self.weight_local
        return w


class LinesLinear(MMMixin, nn.Linear):
    """Linear layer with gloabl and local weights. The weights are interpolated using the
    interpolation constant ``lam``. Thus, the forward pass of this layer will use the interpolated
    weights.

    Note:
        The global weights are the "default" weights of the :class:`torch.nn.Linear` layer,
        while the local ones are in the submodule ``weight_local`` (and ``bias_local``).

    Attributes:
        weight_local (torch.Tensor): The local weights.
        bias_local (torch.Tensor): The local bias.
    """

    def __init__(self, *args, **kwargs: dict[str, Any]):
        super().__init__(*args, **kwargs)
        self.weight_local = nn.Parameter(torch.zeros_like(self.weight))
        if self.bias is not None:
            self.bias_local = nn.Parameter(torch.zeros_like(self.bias))

    def get_weight(self):
        w = (1 - self.lam) * self.weight + self.lam * self.weight_local
        if self.bias is not None:
            b = (1 - self.lam) * self.bias + self.lam * self.bias_local
        else:
            b = None
        return w, b

    def forward(self, x):
        w, b = self.get_weight()
        x = F.linear(input=x, weight=w, bias=b)
        return x


class LinesConv2d(MMMixin, nn.Conv2d):
    """Conv2d layer with gloabl and local weights. The weights are interpolated using the
    interpolation constant ``lam``. Thus, the forward pass of this layer will use the interpolated
    weights.

    Note:
        The global weights are the "default" weights of the :class:`torch.nn.Conv2d` layer, while
        the local ones are in the submodule ``weight_local`` (and ``bias_local``).

    Attributes:
        weight_local (torch.Tensor): The local weights.
        bias_local (torch.Tensor): The local bias.
    """

    def __init__(self, *args, **kwargs: dict[str, Any]):
        super().__init__(*args, **kwargs)
        self.weight_local = nn.Parameter(torch.zeros_like(self.weight))
        if self.bias is not None:
            self.bias_local = nn.Parameter(torch.zeros_like(self.bias))

    def get_weight(self):
        w = (1 - self.lam) * self.weight + self.lam * self.weight_local
        if self.bias is not None:
            b = (1 - self.lam) * self.bias + self.lam * self.bias_local
        else:
            b = None
        return w, b

    def forward(self, x):
        w, b = self.get_weight()
        x = F.conv2d(input=x,
                     weight=w,
                     bias=b,
                     stride=self.stride,
                     padding=self.padding,
                     dilation=self.dilation,
                     groups=self.groups)
        return x


class LinesLSTM(MMMixin, nn.LSTM):
    """LSTM layer with gloabl and local weights. The weights are interpolated using the
    interpolation constant ``lam``. Thus, the forward pass of this layer will use the interpolated
    weights.

    Note:
        The global weights are the "default" weights of the :class:`torch.nn.LSTM` layer, while the
        local ones are in the submodules ``weight_hh_l{layer}_local`` and
        ``weight_ih_l{layer}_local``, where ``layer`` is the layer number. Similar considerations
        apply to the biases.

    Caution:
        This class may not work properly an all devices. If you encounter any issues, please open
        an issue in the repository.


    Attributes:
        weight_hh_l{layer}_local (torch.Tensor): The local hidden-hidden weights of layer ``layer``.
        weight_ih_l{layer}_local (torch.Tensor): The local input-hidden weights of layer ``layer``.
        bias_hh_l{layer}_local (torch.Tensor): The local hidden-hidden biases of layer ``layer``.
        bias_ih_l{layer}_local (torch.Tensor): The local input-hidden biases of layer ``layer``.
    """

    def __init__(self, *args, **kwargs: dict[str, Any]):
        super().__init__(*args, **kwargs)
        for layer in range(self.num_layers):
            setattr(self, f'weight_hh_l{layer}_local', nn.Parameter(
                torch.zeros_like(getattr(self, f'weight_hh_l{layer}'))))
            setattr(self, f'weight_ih_l{layer}_local', nn.Parameter(
                torch.zeros_like(getattr(self, f'weight_ih_l{layer}'))))
            if self.bias:
                setattr(self, f'bias_hh_l{layer}_local', nn.Parameter(
                    torch.zeros_like(getattr(self, f'bias_hh_l{layer}'))))
                setattr(self, f'bias_ih_l{layer}_local', nn.Parameter(
                    torch.zeros_like(getattr(self, f'bias_ih_l{layer}'))))

    def get_weight(self):
        weight_list = []
        for layer in range(self.num_layers):
            weight_list.append((1 - self.lam) * getattr(self,
                               f'weight_ih_l{layer}') +
                               self.lam * getattr(self, f'weight_ih_l{layer}_local'))
            weight_list.append((1 - self.lam) * getattr(self,
                               f'weight_hh_l{layer}') +
                               self.lam * getattr(self, f'weight_hh_l{layer}_local'))
        return weight_list

    def forward(self, x):
        w = self.get_weight()
        h = (
            torch.zeros(self.num_layers, x.shape[0], self.hidden_size).to(x.device),
            torch.zeros(self.num_layers, x.shape[0], self.hidden_size).to(x.device)
        )
        with torch.no_grad():
            # if torch._use_cudnn_rnn_flatten_weight():
            #     torch._cudnn_rnn_flatten_weight(
            #         weight_arr=w,
            #         weight_stride0=(4 if self.bias else 2),
            #         input_size=self.input_size,
            #         mode=2,  # torch.backends.cudnn.rnn.get_cudnn_mode('LSTM'),
            #         hidden_size=self.hidden_size,
            #         proj_size=0,
            #         num_layers=self.num_layers,
            #         batch_first=True,
            #         bidirectional=False
            #     )
            # else:
            self._flat_weights = w
            self.flatten_parameters()
        result = torch._VF.lstm(x, h, w, self.bias, self.num_layers, 0.0,
                                self.training, self.bidirectional, self.batch_first)
        return result[0], result[1:]


class LinesEmbedding(MMMixin, nn.Embedding):
    """Embedding layer with gloabl and local weights. The weights are interpolated using the
    interpolation constant ``lam``. Thus, the forward pass of this layer will use the interpolated
    weights.

    Note:
        The global weights are the "default" weights of the class:`torch.nn.Embedding` layer,
        while the local ones are in the submodule ``weight_local``.

    Attributes:
        weight_local (torch.Tensor): The local weights.
    """

    def __init__(self, *args, **kwargs: dict[str, Any]):
        super().__init__(*args, **kwargs)
        self.weight_local = nn.Parameter(torch.zeros_like(self.weight))

    def forward(self, x):
        w = self.get_weight()
        x = F.embedding(input=x, weight=w)
        return x


class LinesBN2d(MMMixin, nn.BatchNorm2d):
    """BatchNorm2d layer with gloabl and local weights. The weights are interpolated using the
    interpolation constant ``lam``. Thus, the forward pass of this layer will use the interpolated
    weights.

    Note:
        The global weights are the "default" weights of the ``nn.BatchNorm2d`` layer, while the
        local ones are in the submodules ``weight_local`` and ``bias_local``.

    Attributes:
        weight_local (torch.Tensor): The local weights.
        bias_local (torch.Tensor): The local bias.
    """

    def __init__(self, *args, **kwargs: dict[str, Any]):
        super().__init__(*args, **kwargs)
        self.weight_local = nn.Parameter(torch.Tensor(self.num_features))
        self.bias_local = nn.Parameter(torch.Tensor(self.num_features))

    def get_weight(self):
        w = (1 - self.lam) * self.weight + self.lam * self.weight_local
        if self.bias is not None:
            b = (1 - self.lam) * self.bias + self.lam * self.bias_local
        else:
            b = None
        return w, b

    def forward(self, x):
        # call get_weight, which samples from the subspace, then use the corresponding weight.
        w, b = self.get_weight()

        # The rest is code in the PyTorch source forward pass for batchnorm.
        if self.momentum is None:
            exponential_average_factor = 0.0
        else:
            exponential_average_factor = self.momentum

        if self.training and self.track_running_stats:
            if self.num_batches_tracked is not None:
                self.num_batches_tracked = self.num_batches_tracked + 1
                if self.momentum is None:  # use cumulative moving average
                    exponential_average_factor = 1.0 / float(self.num_batches_tracked)
                else:  # use exponential moving average
                    exponential_average_factor = self.momentum

        if self.training:
            bn_training = True
        else:
            bn_training = (self.running_mean is None) and (self.running_var is None)
        return F.batch_norm(
            x,
            self.running_mean if not self.training or self.track_running_stats else None,
            self.running_var if not self.training or self.track_running_stats else None,
            w,
            b,
            bn_training,
            exponential_average_factor,
            self.eps,
        )


def _recursive_mix_networks(merged_net: Module, global_model: Module, local_model: Module) -> dict:
    layers = {}
    if next(merged_net.named_children(), None) is None:
        named_modules = merged_net.named_modules()
    else:
        named_modules = merged_net.named_children()
    for n, x in named_modules:
        if isinstance(x, torch.nn.Linear):
            layer = LinesLinear(x.in_features, x.out_features, bias=x.bias is not None)
        elif isinstance(x, torch.nn.Conv2d):
            layer = LinesConv2d(x.in_channels,
                                x.out_channels,
                                x.kernel_size,
                                x.stride,
                                x.padding,
                                x.dilation,
                                x.groups,
                                x.bias is not None)
        elif isinstance(x, torch.nn.BatchNorm2d):
            layer = LinesBN2d(x.num_features)
        elif isinstance(x, torch.nn.Embedding):
            layer = LinesEmbedding(x.num_embeddings, x.embedding_dim)
        elif isinstance(x, torch.nn.LSTM):
            layer = LinesLSTM(x.input_size,
                              x.hidden_size,
                              x.num_layers,
                              x.bias,
                              x.batch_first,
                              x.dropout,
                              x.bidirectional)
        elif isinstance(x, (torch.nn.BatchNorm1d, torch.nn.BatchNorm3d)):
            raise NotImplementedError("BatchNorm1d and BatchNorm3d are not supported")
        elif next(x.parameters(), None) is None:
            layers[n] = x
            continue
        else:
            layers[n] = _recursive_mix_networks(x,
                                                getattr(global_model, n),
                                                getattr(local_model, n))
            continue

        for namep, _ in x.named_parameters():
            setattr(layer, namep, getattr(global_model, n).get_parameter(
                namep) if n else getattr(global_model, namep))
            setattr(layer, namep + "_local", getattr(local_model, n).get_parameter(namep)
                    if n else getattr(local_model, namep))
        layers[n] = layer

    return layers


def _recursive_set_layer(module: Module, layers: dict):
    for n, l in layers.items():
        if isinstance(l, dict):
            _recursive_set_layer(getattr(module, n), l)
        else:
            setattr(module, n, l)


def mix_networks(global_model: Module, local_model: Module, lamda: float) -> MMMixin:
    """Mix two networks using a linear interpolation.
    This method takes two models and a lambda value and returns a new model that is a linear
    interpolation of the two input models. It transparenly handles the interpolation of the
    different layers of the models. The returned model implements the :class:`MMMixin` class and
    has all the layers swapped with the corresponding interpolated layers.

    See Also:
        - :class:`MMMixin`
        - :class:`LinesLinear`
        - :class:`LinesConv2d`
        - :class:`LinesBN2d`
        - :class:`LinesEmbedding`
        - :class:`LinesLSTM`

    Args:
        global_model (torch.nn.Module): The global model.
        local_model (torch.nn.Module): The local model.
        lamda (float): The interpolation constant.

    Returns:
        Module: The merged/interpolated model that implements the ``MMMixin`` class.
    """
    merged_net = deepcopy(global_model)
    layers = _recursive_mix_networks(merged_net, global_model, local_model)
    if len(layers) == 1 and "" in layers:
        merged_net = layers[""]
    else:
        _recursive_set_layer(merged_net, layers)
    set_lambda_model(merged_net, lamda)
    return merged_net


def _set_lambda(module: MMMixin, lam: float, layerwise: bool = False) -> None:
    """Set model interpolation constant.

    Args:
        module (torch.nn.Module): module
        lam (float): constant used for interpolation (0 means a retrieval of a global model, 1
          means a retrieval of a local model)
        layerwise (bool): set different lambda layerwise or not
    """
    if (
        isinstance(module, torch.nn.Conv2d)
        or isinstance(module, torch.nn.BatchNorm2d)
        or isinstance(module, torch.nn.Linear)
        or isinstance(module, torch.nn.LSTM)
        or isinstance(module, torch.nn.Embedding)
    ):
        if layerwise:
            lam = np.random.uniform(0.0, 1.0)
        setattr(module, 'lam', lam)


def set_lambda_model(model: MMMixin, lam: float, layerwise: bool = False) -> None:
    """Set model interpolation constant.

    Warning:
        This function performs an inplace operation on the model, and
        it assumes that the model has been built using the :class:`MMMixin` classes.

    Args:
        model (torch.nn.Module): model
        lam (float): constant used for interpolation (0 means a retrieval of a global model, 1
          means a retrieval of a local model)
        layerwise (bool): set different lambda layerwise or not
    """
    setattr(model.__class__, 'get_lambda', lambda self: lam)
    model.apply(partial(_set_lambda, lam=lam, layerwise=layerwise))


def get_local_model_dict(model: MMMixin) -> OrderedDict:
    """Get the local model state dictionary.

    Args:
        model (torch.nn.Module): the model.

    Returns:
        OrderedDict: the local model state dictionary.
    """
    return OrderedDict({k.replace("_local", ""): deepcopy(v)
                        for k, v in model.state_dict().items() if "_local" in k})


def get_global_model_dict(model: MMMixin) -> OrderedDict:
    """Get the global model state dictionary.

    Args:
        model (torch.nn.Module): the model.

    Returns:
        OrderedDict: the global model state dictionary.
    """
    return OrderedDict({k: deepcopy(v) for k, v in model.state_dict().items() if "_local" not in k})


def get_output_shape(model: Module, input_dim: tuple[int, ...]) -> tuple[int, ...]:
    """Get the output shape of a model given the shape of the input.

    Args:
        model (torch.nn.Module): The model to get the output shape.
        input_dim (tuple[int, ...]): The expected input shape of the model.

    Returns:
        tuple[int, ...]: The output shape of the model.
    """
    return model(torch.rand(*(input_dim))).data.shape


def diff_model(model_dict1: dict, model_dict2: dict) -> OrderedDict:
    """Compute the difference between two model state dictionaries.
    The difference is computed at the level of the parameters.

    Args:
        model_dict1 (dict): The state dictionary of the first model.
        model_dict2 (dict): The state dictionary of the second model.

    Returns:
        OrderedDict: The state dictionary of the difference between the two models.

    Raises:
        AssertionError: If the two models have different architectures.
    """
    assert model_dict1.keys() == model_dict2.keys(), "Models have not the same architecture"
    return OrderedDict({key: model_dict1[key] - model_dict2[key] for key in model_dict1.keys()})


def merge_models(model_1: Module, model_2: Module, lam: float) -> Module:
    """Merge two models using a linear interpolation.
    The interpolation is done at the level of the parameters using the formula:
    ``merged_model = (1 - lam) * model_1 + lam * model_2``.

    Args:
        model_1 (torch.nn.Module): The first model.
        model_2 (torch.nn.Module): The second model.
        lam (float): The interpolation constant.

    Returns:
        Module: The merged model.
    """
    merged_model = deepcopy(model_1)
    for name, param in merged_model.named_parameters():
        param.data = (1 - lam) * model_1.get_parameter(name).data + \
            lam * model_2.get_parameter(name).data
    return merged_model


def safe_load_state_dict(model1: Module, model2_state_dict: dict) -> None:
    """Load a state dictionary into a model.
    This function is a safe version of ``model.load_state_dict`` that handles the case in which the
    state dictionary has keys that match with :attr:`fluke.utils.model.STATE_DICT_KEYS_TO_IGNORE`
    and thus have to be ignored.

    Caution:
        This function performs an inplace operation on ``model1``.

    Args:
        model1 (torch.nn.Module): The model to load the state dictionary.
        model2_state_dict (dict): The state dictionary.
    """
    model1_state_dict = model1.state_dict()
    new_state_dict = OrderedDict()
    for key, value in model2_state_dict.items():
        if not key.endswith(STATE_DICT_KEYS_TO_IGNORE):
            new_state_dict[key] = value
        else:
            new_state_dict[key] = model1_state_dict[key]
    model1.load_state_dict(new_state_dict)


def batch_norm_to_group_norm(layer: Module) -> Module:
    r"""Iterates over a whole model (or layer of a model) and replaces every
    batch norm 2D with a group norm

    Args:
        layer (torch.nn.Module): model or one layer of a model.

    Returns:
        torch.nn.Module: model with group norm layers instead of batch norm layers.

    Raises:
        ValueError: If the number of channels :math:`\notin \{2^i\}_{i=4}^{11}`
    """

    GROUP_NORM_LOOKUP = {
        16: 2,  # -> channels per group: 8
        32: 4,  # -> channels per group: 8
        64: 8,  # -> channels per group: 8
        128: 8,  # -> channels per group: 16
        256: 16,  # -> channels per group: 16
        512: 32,  # -> channels per group: 16
        1024: 32,  # -> channels per group: 32
        2048: 32,  # -> channels per group: 64
    }

    for name, _ in layer.named_modules():
        if name:
            try:
                sub_layer = getattr(layer, name)
                if isinstance(sub_layer, torch.nn.BatchNorm2d):
                    num_channels = sub_layer.num_features
                    if num_channels in GROUP_NORM_LOOKUP:
                        layer._modules[name] = torch.nn.GroupNorm(
                            GROUP_NORM_LOOKUP[num_channels], num_channels)
                    else:
                        raise ValueError(
                            f"GroupNorm not implemented for {num_channels} channels")
            except AttributeError:
                name = name.split('.')[0]
                sub_layer = getattr(layer, name)
                sub_layer = batch_norm_to_group_norm(sub_layer)
                layer.__setattr__(name=name, value=sub_layer)
    return layer


def state_dict_zero_like(state_dict: OrderedDict) -> OrderedDict:
    """Create a state dictionary with the same keys as the input state dictionary but with zeros
    tensors.

    Args:
        state_dict (OrderedDict): The state dictionary.

    Returns:
        OrderedDict: The state dictionary with zeros tensors.
    """
    output = OrderedDict()
    for k, v in state_dict.items():
        output[k] = torch.zeros_like(v)
    return output


def flatten_parameters(model: torch.nn.Module) -> torch.Tensor:
    """Returns the model parameters as a contiguous tensor.

    Args:
        model (torch.nn.Module): The model.

    Returns:
        torch.Tensor: The model parameters as a contiguous tensor of shape (n,), where n is the
            number of parameters in the model.
    """
    n = sum(p.numel() for p in model.parameters())
    params = torch.zeros(n)
    i = 0
    for p in model.parameters():
        params_slice = params[i:i + p.numel()]
        params_slice.copy_(p.flatten())
        p.data = params_slice.view(p.shape)
        i += p.numel()
    return params


def get_activation_size(model: nn.Module, input_tensor: torch.Tensor = None) -> int:
    """Get the size of the activations of the model.
    This method computes the size of the activations of the model given the input tensor. If the
    model has a linear layer as the last layer, the size of the activations is the number of
    features of the linear layer (thus, the input tensor is not requires).
    If the model has a different last layer, the method computes the
    size of the activations by performing a forward pass with the input tensor.

    Args:
        model (nn.Module): The model.
        input_tensor (torch.Tensor, optional): The input tensor. Defaults to `None`.

    Raises:
        ValueError: If the input tensor is `None` and the last layer of the model is not a linear
            layer.

    Returns:
        int: The size of the activations of the model.
    """
    last_layer = list(model.modules())[-1]
    if isinstance(last_layer, nn.Linear):
        return last_layer.out_features
    elif input_tensor is None:
        raise ValueError(
            "'input_tensor' must be not None if the last layer of the network is not nn.Linear.")
    return model(input_tensor).numel()


def get_trainable_keys(model: nn.Module) -> list[str]:
    """Get the keys of the model parameters that are trainable (i.e., require gradients).

    Args:
        model (nn.Module): The model.

    Returns:
        list[str]: The keys of the model parameters that are trainable.
    """
    return [k for k, p in model.named_parameters() if p.requires_grad]


def aggregate_models(target_model: nn.Module,
                     models: Iterable[nn.Module],
                     weights: Iterable[float],
                     eta: float,
                     inplace: bool = True) -> nn.Module:
    r"""Aggregate the models using a weighted average.
    The method aggregates the models using a weighted average of the parameters and updates the
    target model with the aggregated parameters. The aggregation is done using the formula:

    .. math::

        \theta_{t+1} = (1 - \eta) \theta_t + \eta \sum_{i=1}^{N} w_i \theta_i

    where :math:`\theta_t` is the target model, :math:`\theta_i` are the models to aggregate,
    :math:`w_i` are the weights of the models, :math:`N` is the number of models, and :math:`\eta`
    is the learning rate of the aggregation.

    Args:
        target_model (nn.Module): The target model.
        models (list[nn.Module]): The models to aggregate.
        weights (list[float]): The weights of the models.
        eta (float): The learning rate of the aggregation.
        inplace (bool, optional): If ``True``, the target model is updated in place. Defaults to
            ``True``.

    Returns:
        nn.Module: The updated target model.
    """

    if not inplace:
        target_model = deepcopy(target_model)

    # Get model parameters and buffers
    model_params = dict(target_model.named_parameters())
    model_buffers = dict(target_model.named_buffers())  # Includes running_mean, running_var, etc.

    # Initialize accumulators for parameters
    avg_params = {key: torch.zeros_like(param.data) for key, param in model_params.items()}
    avg_buffers = {key: torch.zeros_like(buffer.data)
                   for key, buffer in model_buffers.items() if "num_batches_tracked" not in key}

    max_num_batches_tracked = 0  # Track the max num_batches_tracked

    # Compute weighted sum (weights already sum to 1, so no division needed)
    for m, w in zip(models, weights):
        for key, param in m.named_parameters():
            avg_params[key].add_(param.data, alpha=w)

        for key, buffer in m.named_buffers():
            if "num_batches_tracked" not in key:
                avg_buffers[key].add_(buffer.data, alpha=w)
            else:
                max_num_batches_tracked = max(max_num_batches_tracked, buffer.item())

    for key in model_params.keys():
        model_params[key].data.lerp_(avg_params[key], eta)  # Soft update

    for key in model_buffers.keys():
        if "num_batches_tracked" not in key:
            model_buffers[key].data.lerp_(avg_buffers[key], eta)

    # Assign max num_batches_tracked
    for key in model_buffers.keys():
        if "num_batches_tracked" in key:
            model_buffers[key].data.fill_(max_num_batches_tracked)

    return target_model


def check_model_fit_mem(model: torch.nn.Module,
                        input_size: tuple[int, ...],
                        num_clients: int,
                        device: str = 'cuda',
                        mps_default: bool = True):
    """Check if the models fit in the memory of the device.
    The method estimates the memory usage of the models, when all clients and the server own a
    single neural network, on the device and checks if the models fit in the memory of the device.

    Attention:
        This function only works for CUDA devices. For MPS devices, the function will
        always return the value of ``mps_default``. To date, PyTorch does not provide
        a way to estimate the memory usage of a model on an MPS device.

    Args:
        model (torch.nn.Module): The model to check.
        input_size (tuple[int, ...]): The input size of the model.
        num_clients (int): The number of clients in the federation.
        device (str, optional): The device to check. Defaults to 'cuda'.
        mps_default (bool, optional): The default value to return if the device is MPS.
    """

    # Ensure the device is available
    assert device in ["cuda", "mps"] or device.startswith("cuda:"), \
        "Invalid argument 'device'. Must be 'cuda', 'mps' or 'cuda:<device_id>'."

    if device == "mps":
        if not torch.backends.mps.is_available():
            raise RuntimeError("MPS is not available.")
    elif not torch.cuda.is_available():
        raise RuntimeError(f"CUDA is not available on device: {device}")

    if device == "mps":
        return mps_default
    else:
        return _check_model_fit_mem_cuda(model, input_size, num_clients, device)


def _check_model_fit_mem_cuda(model: torch.nn.Module,
                              input_size: tuple[int, ...],
                              num_clients: int,
                              device: torch.device):

    # Get the current CUDA device
    cuda_device = torch.device(device)

    # Get the free memory in bytes
    free_mem = torch.cuda.mem_get_info(cuda_device.index)[0]
    # current_allocation = torch.cuda.memory_allocated(cuda_device)
    current_reserved = torch.cuda.memory_reserved(cuda_device)

    # Transfer the model to CUDA
    model = model.to(cuda_device)

    # Estimate the model memory
    # param_memory = sum(p.numel() * p.element_size() for p in model.parameters())

    # Create a dummy input tensor with the specified size
    dummy_input = torch.randn(*input_size, device=cuda_device)

    # Forward pass to estimate memory for activations
    with torch.no_grad():
        model(dummy_input)

    # Get the current GPU memory allocated and cached
    # allocated_mem = torch.cuda.memory_allocated(cuda_device)
    reserved_mem = torch.cuda.memory_reserved(cuda_device)

    # Total estimated memory usage
    total_estimated_mem = (num_clients + 1) * (reserved_mem - current_reserved)

    # Compare the total estimated memory with the free memory
    return total_estimated_mem <= free_mem


def _recursive_register_hook(module: Module, hook: callable, name: str = "", handles: list = None):
    named_modules = module.named_children()
    empty = True
    if handles is None:
        handles = []
    for n, sub_module in named_modules:
        empty = False
        current_name = name + "." + n if name else n
        leaf = _recursive_register_hook(sub_module, hook, current_name, handles)
        if leaf:
            handles.append(sub_module.register_forward_hook(hook(current_name)))
    return empty


class AllLayerOutputModel(nn.Module):
    """Wrapper class to get the output of all layers in a model.
    Once the model is wrapped with this class, the activations of all layers can be accessed through
    the attributes ``activations_in`` and ``activations_out``.

    ``activations_in`` is a dictionary that contains the input activations of all layers.
    ``activations_out`` is a dictionary that contains the output activations of all layers.

    Note:
        The activations are stored in the order in which they are computed during the forward pass.

    Important:
        If you need to access the activations of a specific layer after a potential activation
        function, you should use the ``activations_in`` of the next layer. For example, if you
        need the activations of the first layer after the ReLU activation, you should use the
        activations_in of the second layer. These attribute may not include the activations of the
        last layer if it includes an activation function.

    Important:
        If your model includes as submodule all the activations functions (e.g., of type
        torch.nn.ReLU), then you can use the ``activations_out`` attribute to get all the
        activations (i.e., before and after the activation functions).

    Attributes:
        model (torch.nn.Module): The model to wrap.
        activations_in (OrderedDict): The input activations of all layers.
        activations_out (OrderedDict): The output activations of all layers.

    Args:
        model (torch.nn.Module): The model to wrap.
    """

    def __init__(self, model: nn.Module):
        super().__init__()
        self.model = model
        self.activations_in = OrderedDict()
        self.activations_out = OrderedDict()
        self._handles = []
        self.activate()

    def is_active(self) -> bool:
        """Returns whether the all layer output model is active.

        Returns:
            bool: Whether the all layer output model is active.
        """
        return bool(self._handles)

    def activate(self) -> None:
        """Activate the all layer output functionality."""
        _recursive_register_hook(self.model, self._get_activation, handles=self._handles)
        # for layer in self.model.modules():
        #     self._handles.append(layer.register_forward_hook(self._get_activation(layer)))

    def deactivate(self, clear_activations: bool = True) -> None:
        """Deactivate the all layer output functionality.

        Args:
            clear_activations (bool, optional): Whether to clear the stored activations. Defaults to
                ``True``.
        """
        for h in self._handles:
            h.remove()
        self._handles = []
        if clear_activations:
            self.activations_in = OrderedDict()
            self.activations_out = OrderedDict()

    def _get_activation(self, name):
        def hook(model, input, output):
            # if name not in self.activations_in:
            self.activations_in[name] = input[0].detach()
            self.activations_out[name] = output.detach()
        return hook

    def forward(self, x):
        return self.model(x)



"""Implementation of the [FedPer19]_ algorithm.

References:
    .. [FedPer19] Manoj Ghuhan Arivazhagan, Vinay Aggarwal, Aaditya Kumar Singh, and
       Sunav Choudhary. Federated learning with personalization layers.
       In arXiv (2019). URL:https://arxiv.org/abs/1912.00818
"""
import sys
from copy import deepcopy
from typing import Any, Iterable

import torch

sys.path.append(".")
sys.path.append("..")

from ..algorithms import PersonalizedFL  # NOQA
from ..client import Client  # NOQA
from ..comm import Message  # NOQA
from ..data import FastDataLoader  # NOQA
from ..nets import EncoderGlobalHeadLocalNet, EncoderHeadNet  # NOQA
from ..server import Server  # NOQA
from ..utils import OptimizerConfigurator  # NOQA
from ..utils.model import safe_load_state_dict  # NOQA


# https://arxiv.org/abs/1912.00818
class FedPerClient(Client):

    def __init__(self,
                 index: int,
                 model: EncoderHeadNet,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: torch.nn.Module,
                 local_epochs: int = 3,
                 fine_tuning_epochs: int = 0,
                 clipping: float = 0,
                 **kwargs: dict[str, Any]):
        super().__init__(index=index, train_set=train_set, test_set=test_set,
                         optimizer_cfg=optimizer_cfg, loss_fn=loss_fn, local_epochs=local_epochs,
                         fine_tuning_epochs=fine_tuning_epochs, clipping=clipping, **kwargs)
        self.model = EncoderGlobalHeadLocalNet(model)
        self._save_to_cache()

    def send_model(self):
        self.channel.send(Message(deepcopy(self.model.get_global()), "model", self, inmemory=True),
                          self.server)

    def receive_model(self) -> None:
        msg = self.channel.receive(self, self.server, msg_type="model")
        safe_load_state_dict(self.model.get_global(), msg.payload.state_dict())


class FedPerServer(Server):

    def __init__(self,
                 model: torch.nn.Module,
                 test_set: FastDataLoader,  # not used
                 clients: Iterable[Client],
                 weighted: bool = False):
        super().__init__(model=model, test_set=None, clients=clients, weighted=weighted)


class FedPer(PersonalizedFL):

    def get_client_class(self) -> FedPerClient:
        return FedPerClient

    def get_server_class(self) -> Server:
        return FedPerServer



"""Implementation of the [Per-FedAVG20]_ algorithm.

References:
    .. [Per-FedAVG20] Alireza Fallah, Aryan Mokhtari, Asuman Ozdaglar. Personalized Federated
       Learning with Theoretical Guarantees: A Model-Agnostic Meta-Learning Approach.
       In NeurIPS (2020). URL: https://arxiv.org/abs/2002.07948
"""
import sys
from collections import OrderedDict
from copy import deepcopy
from typing import Any, Union

import torch
from torch.optim import Optimizer

sys.path.append(".")
sys.path.append("..")

from ..algorithms import CentralizedFL  # NOQA
from ..client import Client  # NOQA
from ..data import FastDataLoader  # NOQA
from ..utils import OptimizerConfigurator  # NOQA

__all__ = [
    "PerFedAVGOptimizer",
    "PerFedAVGClient",
    "PerFedAVG"
]


class PerFedAVGOptimizer(Optimizer):
    def __init__(self, params, lr):
        defaults = dict(lr=lr)
        super(PerFedAVGOptimizer, self).__init__(params, defaults)

    def step(self, closure=None, beta=0, grads=None):
        loss = None
        if closure is not None:
            loss = closure

        for group in self.param_groups:
            if grads is None:
                for p in group['params']:
                    if p.grad is None:
                        continue
                    d_p = p.grad.data
                    p.data.sub_(d_p, alpha=beta if (beta != 0) else group['lr'])
            else:
                for p, g1, g2 in zip(group['params'], grads[0], grads[1]):
                    if p.grad is None:
                        continue
                    p.data.sub_(beta * g1 - beta * group['lr'] * g2)
        return loss


class PerFedAVGClient(Client):
    def __init__(self,
                 index: int,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: torch.nn.Module,
                 local_epochs: int,
                 mode: str,
                 beta: float,
                 fine_tuning_epochs: int = 0,
                 clipping: float = 0,
                 **kwargs: dict[str, Any]):

        super().__init__(index=index, train_set=train_set, test_set=test_set,
                         optimizer_cfg=optimizer_cfg, loss_fn=loss_fn, local_epochs=local_epochs,
                         fine_tuning_epochs=fine_tuning_epochs, clipping=clipping, **kwargs)
        self.hyper_params.update(
            mode=mode,
            beta=beta
        )
        self.train_iterator = iter(self.train_set)

    def _get_next_batch(self):
        try:
            X, y = next(self.train_set)
        except StopIteration:
            self.train_iterator = iter(self.train_set)
            X, y = next(self.train_iterator)

        return X, y

    def _compute_grad(self,
                      model: torch.nn.Module,
                      data_batch: tuple[torch.Tensor, torch.Tensor],
                      v: Union[tuple[torch.Tensor, ...], None] = None) -> list[torch.Tensor]:
        X, y = data_batch
        X, y = X.to(self.device), y.to(self.device)
        if v is not None:
            frz_model_params = deepcopy(model.state_dict())
            delta = 1e-3
            dummy_model_params_1 = OrderedDict()
            dummy_model_params_2 = OrderedDict()
            with torch.no_grad():
                for (layer_name, param), grad in zip(model.named_parameters(), v):
                    dummy_model_params_1.update({layer_name: param + delta * grad})
                    dummy_model_params_2.update({layer_name: param - delta * grad})

            model.load_state_dict(dummy_model_params_1, strict=False)
            logit_1 = model(X)
            loss_1 = self.hyper_params.loss_fn(logit_1, y)
            grads_1 = torch.autograd.grad(loss_1, model.parameters())

            model.load_state_dict(dummy_model_params_2, strict=False)
            logit_2 = model(X)
            loss_2 = self.hyper_params.loss_fn(logit_2, y)
            grads_2 = torch.autograd.grad(loss_2, model.parameters())

            model.load_state_dict(frz_model_params)

            grads = []
            with torch.no_grad():
                for g1, g2 in zip(grads_1, grads_2):
                    grads.append((g1 - g2) / (2 * delta))
            return grads
        else:
            logit = model(X)
            loss = self.hyper_params.loss_fn(logit, y)
            grads = torch.autograd.grad(loss, model.parameters())
            return grads

    def fit(self, override_local_epochs: int = 0) -> float:
        epochs: int = (override_local_epochs if override_local_epochs > 0
                       else self.hyper_params.local_epochs)
        self.model.train()
        self.model.to(self.device)
        if self.optimizer is None:
            self.optimizer, _ = self._optimizer_cfg(self.model)

        iterations = len(self.train_set) * epochs
        running_loss = 0.0
        for _ in range(iterations):
            batch_1 = self._get_next_batch()
            batch_2 = self._get_next_batch()

            X, y = batch_1
            X, y = X.to(self.device), y.to(self.device)
            self.optimizer.zero_grad()
            y_hat = self.model(X)
            loss = self.hyper_params.loss_fn(y_hat, y)
            loss.backward()
            self.optimizer.step(self.model.parameters())
            running_loss += loss.item()

            if self.hyper_params.mode == "FO":
                X, y = batch_2
                X, y = X.to(self.device), y.to(self.device)
                self.optimizer.zero_grad()
                y_hat = self.model(X)
                loss = self.hyper_params.loss_fn(y_hat, y)
                loss.backward()
                self._clip_grads(self.model)
                self.optimizer.step(self.model.parameters(), beta=self.hyper_params.beta)

            elif self.hyper_params.mode == "HF":
                batch_3 = self._get_next_batch()

                temp_model = deepcopy(self.model).to(self.device)
                grads_1st = self._compute_grad(temp_model, batch_2)
                grads_2nd = self._compute_grad(self.model, batch_3, v=grads_1st)
                self.optimizer.step(self.model.parameters(),
                                    beta=self.hyper_params.beta,
                                    grads=(grads_1st, grads_2nd))

            else:
                raise ValueError(f"Invalid mode: {self.hyper_params.mode}")

        running_loss /= iterations
        self.model.cpu()
        return running_loss


class PerFedAVG(CentralizedFL):

    def get_client_class(self) -> Client:
        return PerFedAVGClient

    def can_override_optimizer(self) -> bool:
        return False

    def get_optimizer_class(self) -> Optimizer:
        return PerFedAVGOptimizer



"""Implementation of the [SCAFFOLD20]_ algorithm.

References:
    .. [SCAFFOLD20] Sai Praneeth Karimireddy, Satyen Kale, Mehryar Mohri, Sashank J. Reddi,
       Sebastian U. Stich, Ananda Theertha Suresh. SCAFFOLD: Stochastic Controlled Averaging for
       Federated Learning. In ICML (2020). URL: https://arxiv.org/abs/1910.06378
"""
import sys
from collections import OrderedDict
from copy import deepcopy
from typing import Any, Iterable

import torch
from torch.nn import Module

sys.path.append(".")
sys.path.append("..")

from .. import FlukeENV  # NOQA
from ..algorithms import CentralizedFL  # NOQA
from ..client import Client  # NOQA
from ..comm import Message  # NOQA
from ..data import FastDataLoader  # NOQA
from ..server import Server  # NOQA
from ..utils import OptimizerConfigurator, clear_cuda_cache  # NOQA
from ..utils.model import safe_load_state_dict, state_dict_zero_like  # NOQA

__all__ = [
    "SCAFFOLDClient",
    "SCAFFOLDServer",
    "SCAFFOLD"
]


class SCAFFOLDClient(Client):
    def __init__(self,
                 index: int,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: torch.nn.Module,
                 local_epochs: int = 3,
                 fine_tuning_epochs: int = 0,
                 clipping: float = 0,
                 **kwargs: dict[str, Any]):
        super().__init__(index=index, train_set=train_set, test_set=test_set,
                         optimizer_cfg=optimizer_cfg, loss_fn=loss_fn, local_epochs=local_epochs,
                         fine_tuning_epochs=fine_tuning_epochs, clipping=clipping, **kwargs)
        self.control: OrderedDict = None
        self.delta_control: OrderedDict = None
        self.server_control: OrderedDict = None
        self._attr_to_cache.extend(["control", "delta_control", "server_control"])

    def receive_model(self) -> None:
        model = self.channel.receive(self, self.server, msg_type="model").payload
        self.server_control = self.channel.receive(self, self.server, msg_type="control").payload
        if self.model is None:
            self.model = model
            self.control = state_dict_zero_like(model.state_dict())
        else:
            safe_load_state_dict(self.model, model.state_dict())
        self.server_model = deepcopy(model.state_dict())

    def fit(self, override_local_epochs: int = 0) -> float:
        epochs: int = (override_local_epochs if override_local_epochs > 0
                       else self.hyper_params.local_epochs)
        self.model.to(self.device)
        self.model.train()

        if self.optimizer is None:
            self.optimizer, self.scheduler = self._optimizer_cfg(self.model)

        K = 0
        running_loss = 0.0
        for _ in range(epochs):
            loss = None
            for _, (X, y) in enumerate(self.train_set):
                K += 1
                X, y = X.to(self.device), y.to(self.device)
                self.optimizer.zero_grad()
                y_hat = self.model(X)
                loss = self.hyper_params.loss_fn(y_hat, y)
                loss.backward()

                for n, p in self.model.named_parameters():
                    p.grad.data = p.grad.data + (self.server_control[n].to(self.device) -
                                                 self.control[n].to(self.device))

                self._clip_grads(self.model)
                self.optimizer.step()
                running_loss += loss.item()
            self.scheduler.step()

        self.model.cpu()
        # This happens on CPU!

        with torch.no_grad():
            c_plus = state_dict_zero_like(self.control)
            c_delta = state_dict_zero_like(self.control)
            model_params = self.model.state_dict()
            for key in model_params:
                c_plus[key] = self.control[key] - self.server_control[key] + \
                    (self.server_model[key] - model_params[key]) / \
                    (K * self.scheduler.get_last_lr()[0])

            for key in model_params:
                c_delta[key] = c_plus[key] - self.control[key]

            self.control = deepcopy(c_plus)
            self.delta_control = c_delta

        running_loss /= (epochs * len(self.train_set))
        clear_cuda_cache()
        return running_loss

    def send_model(self):
        self.channel.send(Message(self.model, "model", self, inmemory=True), self.server)
        self.channel.send(Message(self.delta_control, "control", self, inmemory=True), self.server)


class SCAFFOLDServer(Server):
    def __init__(self,
                 model: Module,
                 test_set: FastDataLoader,
                 clients: Iterable[Client],
                 weighted: bool = True,
                 global_step: float = 1.,
                 **kwargs: dict[str, Any]):
        super().__init__(model=model,
                         test_set=test_set,
                         clients=clients,
                         weighted=weighted,
                         **kwargs)
        self.device = FlukeENV().get_device()
        self.control = state_dict_zero_like(self.model.state_dict())
        self.hyper_params.update(global_step=global_step)

    def broadcast_model(self, eligible: Iterable[Client]) -> None:
        self.channel.broadcast(Message(self.model, "model", self), eligible)
        self.channel.broadcast(Message(self.control, "control", self), eligible)

    def _get_client_weights(self, eligible: Iterable[Client]):
        weights = super()._get_client_weights(eligible)
        if self.hyper_params.global_step != 1:
            for c in range(len(eligible)):
                weights[c] = self.hyper_params.global_step * weights[c]
        return weights

    @torch.no_grad()
    def aggregate(self, eligible: Iterable[Client], client_models: Iterable[Module]) -> None:
        self.model.to(self.device)

        total_delta = state_dict_zero_like(self.model.state_dict())
        for client in eligible:
            c_delta_params = self.channel.receive(self, client, "control").payload
            for key in total_delta:
                total_delta[key] = total_delta[key] + c_delta_params[key].to(self.device)

        for key in total_delta:
            total_delta[key] = total_delta[key] / self.n_clients
            self.control[key] = self.control[key] + total_delta[key].to("cpu")

        self.model.cpu()
        return super().aggregate(eligible, client_models)


class SCAFFOLD(CentralizedFL):

    def get_client_class(self) -> Client:
        return SCAFFOLDClient

    def get_server_class(self) -> Server:
        return SCAFFOLDServer



import gc
import sys
import tempfile
from typing import Any
from torch.nn import CrossEntropyLoss, Module
from torch.optim import SGD
import numpy as np
sys.path.append(".")
sys.path.append("..")

from fluke import DDict, FlukeENV  # NOQA
from fluke.algorithms import CentralizedFL, PersonalizedFL  # NOQA
from fluke.client import Client  # NOQA
from fluke.comm import ChannelObserver, Message  # NOQA
from fluke.data import DataSplitter  # NOQA
from fluke.data.datasets import Datasets  # NOQA
from fluke.evaluation import ClassificationEval  # NOQA
from fluke.nets import MNIST_2NN  # NOQA
from fluke.server import Server  # NOQA
from fluke.utils import (ClientObserver, Configuration, ServerObserver,  # NOQA
                         get_class_from_qualified_name)
from fluke.utils.log import Log  # NOQA

FlukeENV().set_evaluator(ClassificationEval(1, 10))
FlukeENV().set_eval_cfg(DDict(post_fit=True, pre_fit=True))


def test_centralized_fl():
    hparams = DDict(
        # model="fluke.nets.MNIST_2NN",
        model=MNIST_2NN(),
        client=DDict(batch_size=32,
                     local_epochs=1,
                     loss=CrossEntropyLoss,
                     optimizer=DDict(
                         lr=0.1,
                         momentum=0.9),
                     scheduler=DDict(
                         step_size=1,
                         gamma=0.1)
                     ),
        server=DDict(weighted=True)
    )
    mnist = Datasets.MNIST("../data")
    splitter = DataSplitter(mnist, client_split=0.1)
    fl = CentralizedFL(2, splitter, hparams)

    assert fl.n_clients == 2
    assert fl.hyper_params == hparams
    assert len(fl.clients) == 2
    assert isinstance(fl.clients[0], Client)
    assert isinstance(fl.server, Server)
    assert isinstance(fl.server.model, MNIST_2NN)
    assert fl.clients[0].test_set is not None
    assert fl.clients[1].test_set is not None
    assert isinstance(fl.clients[0].hyper_params.loss_fn, CrossEntropyLoss)

    hparams = DDict(
        model="fluke.nets.MNIST_2NN",
        client=DDict(batch_size=32,
                     local_epochs=1,
                     loss="CrossEntropyLoss",
                     optimizer=DDict(
                         lr=0.1,
                         momentum=0.9),
                     scheduler=DDict(
                         step_size=1,
                         gamma=0.1)
                     ),
        server=DDict(weighted=True)
    )
    mnist = Datasets.MNIST("../data")
    splitter = DataSplitter(mnist, client_split=0.1)
    fl = CentralizedFL(2, splitter, hparams)

    assert isinstance(fl.server.model, MNIST_2NN)
    assert isinstance(fl.clients[0].hyper_params.loss_fn, CrossEntropyLoss)

    assert fl.get_optimizer_class() == SGD
    assert fl.get_client_class() == Client
    assert fl.get_server_class() == Server

    class Observer(ClientObserver, ServerObserver, ChannelObserver):
        def __init__(self):
            super().__init__()
            self.called_start = False
            self.called_end = False
            self.called_selected = False
            self.called_error = False
            self.called_finished = False
            self.called_client_eval = False
            self.called_server_eval = False
            self.called_start_fit = False
            self.called_end_fit = False

        def start_round(self, round, global_model):
            assert round == 1
            assert global_model is not None
            self.called_start = True

        def end_round(self, round):
            assert round == 1
            self.called_end = True

        def server_evaluation(self, round, type, evals) -> None:
            assert round == 1
            assert type == "global"
            assert "accuracy" in evals
            self.called_server_eval = True

        def client_evaluation(self, round, client_id, phase, evals, **kwargs: dict[str, Any]):
            assert round == 1 or (round == -1 and phase == "pre-fit")
            assert phase == "post-fit" or phase == "pre-fit"
            assert client_id == 0 or client_id == 1
            self.called_client_eval = True

        def selected_clients(self, round, clients):
            assert len(clients) == 1
            assert round == 1
            self.called_selected = True

        def finished(self, round):
            assert round == 2
            self.called_finished = True

        def start_fit(self, round: int, client_id: int, model: Module, **kwargs: dict[str, Any]):
            assert round == 1
            assert client_id == 0 or client_id == 1
            self.called_start_fit = True

        def end_fit(self, round: int, client_id: int, model: Module, loss: float, **kwargs: dict[str, Any]):
            assert round == 1
            assert client_id == 0 or client_id == 1
            assert loss >= 0.0
            self.called_end_fit = True

        def message_received(self, message: Message):
            pass

    obs = Observer()
    fl.set_callbacks(obs)

    assert fl.server._observers == [obs]

    strfl = f"CentralizedFL[{fl.id}](model=fluke.nets.MNIST_2NN,Client[0-1](optim=OptCfg(SGD,lr=0.1," + \
        "momentum=0.9,StepLR(step_size=1,gamma=0.1)),batch_size=32,loss_fn=CrossEntropyLoss()," + \
        "local_epochs=1,fine_tuning_epochs=0,clipping=0),Server(weighted=True,lr=1.0))"

    assert str(fl).replace(" ", "").replace(
        "\n", "").replace("\t", "") == strfl
    assert fl.__repr__().replace(" ", "").replace(
        "\n", "").replace("\t", "") == strfl

    fl.run(1, 0.5)
    assert obs.called_start
    assert obs.called_end
    assert obs.called_selected
    assert obs.called_finished
    assert obs.called_start_fit
    assert obs.called_end_fit
    assert obs.called_client_eval
    assert obs.called_server_eval

    with tempfile.TemporaryDirectory() as tmpdirname:
        temppath = fl.save(tmpdirname)
        fl2 = CentralizedFL(2, splitter, hparams)
        fl2.load(temppath)

        assert fl2.server.rounds == fl.server.rounds

    FlukeENV().set_inmemory(True)
    hparams = DDict(
        # model="fluke.nets.MNIST_2NN",
        model=MNIST_2NN(),
        client=DDict(batch_size=32,
                     local_epochs=1,
                     model=MNIST_2NN(),
                     loss=CrossEntropyLoss,
                     optimizer=DDict(
                         lr=0.1,
                         momentum=0.9),
                     scheduler=DDict(
                         step_size=1,
                         gamma=0.1)
                     ),
        server=DDict(weighted=True)
    )
    mnist = Datasets.MNIST("../data")
    splitter = DataSplitter(mnist)
    fl = PersonalizedFL(2, splitter, hparams)

    assert fl.n_clients == 2
    assert fl.hyper_params == hparams
    assert len(fl.clients) == 2
    assert isinstance(fl.clients[0], Client)
    assert isinstance(fl.server, Server)
    assert isinstance(fl.server.model, MNIST_2NN)
    # assert isinstance(fl.clients[0].model, MNIST_2NN)
    assert fl.clients[0].test_set is None
    assert isinstance(fl.clients[0].hyper_params.loss_fn, CrossEntropyLoss)

    fl.run(1, 0.5)
    with tempfile.TemporaryDirectory() as tmpdirname:
        temppath = fl.save(tmpdirname)
        fl2 = PersonalizedFL(2, splitter, hparams)
        fl2.load(temppath)

        assert fl2.server.rounds == fl.server.rounds


def get_splitter(cfg):
    dataset = Datasets.get(**cfg.data.dataset)
    splitter = DataSplitter(dataset=dataset,
                            distribution=cfg.data.distribution.name,
                            dist_args=cfg.data.distribution.exclude("name"),
                            **cfg.data.exclude('dataset', 'distribution'))
    return splitter


SPLITTER = None


def _test_algo(exp_config, alg_config, rounds=2, oncpu=True):
    accs = []
    cfg = Configuration(exp_config, alg_config)
    if oncpu:
        cfg.exp.device = "cpu"
    else:
        # if torch.cuda.is_available():
        #     cfg.exp.device = "cuda"
        # elif torch.backends.mps.is_available():
        #     cfg.exp.device = "mps"
        # else:
        return None, None
    FlukeENV().configure(cfg)
    gc.collect()
    global SPLITTER
    if SPLITTER is None:
        dataset = Datasets.get(**cfg.data.dataset)
        SPLITTER = DataSplitter(dataset=dataset,
                                distribution=cfg.data.distribution.name,
                                dist_args=cfg.data.distribution.exclude("name"),
                                **cfg.data.exclude('dataset', 'distribution'))
    fl_algo_class = get_class_from_qualified_name(cfg.method.name)

    for mem in [False, True]:
        FlukeENV().configure(cfg)
        cfg.exp.inmemory = mem
        FlukeENV().set_inmemory(mem)
        algo = fl_algo_class(cfg.protocol.n_clients,
                             SPLITTER,
                             cfg.method.hyperparameters)

        log = Log()
        log.init(**cfg)
        algo.set_callbacks(log)
        algo.run(cfg.protocol.n_rounds, cfg.protocol.eligible_perc)
        FlukeENV().close_cache()
        del algo
        if log.global_eval:
            accs.append(log.global_eval[cfg.protocol.n_rounds]["accuracy"])
        else:
            accs.append(log.postfit_eval_summary[cfg.protocol.n_rounds]["accuracy"])

    assert np.allclose(accs[0], accs[1], atol=1e-5)
    return None, log


def test_apfl():
    apfl, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/apfl.yaml")
    # apfl, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/apfl.yaml", oncpu=False)


def test_ccvr():
    ccvr, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/ccvr.yaml")
    # ccvr, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/ccvr.yaml", oncpu=False)


def test_ditto():
    ditto, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/ditto.yaml")
    # ditto, log = _test_algo("./tests/configs/exp.yaml",
    #                         "./tests/configs/alg/ditto.yaml", oncpu=False)


def test_fat():
    fat, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fat.yaml")


def test_fedala():
    fedala, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedala.yaml")
    # fedala, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/fedala.yaml", oncpu=False)


def test_fedamp():
    fedamp, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedamp.yaml")
    # fedamp, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/fedamp.yaml", oncpu=False)


def test_fedavg():
    fedavg, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedavg.yaml")
    # fedavg, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/fedavg.yaml", oncpu=False)


def test_fedavgm():
    fedavgm, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedavgm.yaml")
    # fedavgm, log = _test_algo("./tests/configs/exp.yaml",
    #                           "./tests/configs/alg/fedavgm.yaml", oncpu=False)


def test_fedaws():
    fedaws, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedaws.yaml")
    # fedaws, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/fedaws.yaml", oncpu=False)


def test_fedbabu():
    fedbabu, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedbabu.yaml")
    fedbabu, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedbabu_head.yaml")
    fedbabu, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedbabu_body.yaml")
    # fedbabu, log = _test_algo("./tests/configs/exp.yaml",
    #                           "./tests/configs/alg/fedbabu.yaml", oncpu=False)
    # fedbabu, log = _test_algo("./tests/configs/exp.yaml",
    #                           "./tests/configs/alg/fedbabu_head.yaml", oncpu=False)
    # fedbabu, log = _test_algo("./tests/configs/exp.yaml",
    #                           "./tests/configs/alg/fedbabu_body.yaml", oncpu=False)


def test_fedbn():
    fedbn, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedbn.yaml")
    # fedbn, log = _test_algo("./tests/configs/exp.yaml",
    #                         "./tests/configs/alg/fedbn.yaml", oncpu=False)


def test_feddyn():
    feddyn, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/feddyn.yaml")
    # feddyn, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/feddyn.yaml", oncpu=False)


def test_fedexp():
    fedexp, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedexp.yaml")
    # fedexp, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/fedexp.yaml", oncpu=False)


def test_fedhp():
    fedhp, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedhp.yaml")
    # fedhp, log = _test_algo("./tests/configs/exp.yaml",
    #                         "./tests/configs/alg/fedhp.yaml", oncpu=False)


def test_fedlc():
    fedlc, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedlc.yaml")
    # fedlc, log = _test_algo("./tests/configs/exp.yaml",
    #                         "./tests/configs/alg/fedlc.yaml", oncpu=False)


def test_fedlc():
    fedlc, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedld.yaml")
    # fedlc, log = _test_algo("./tests/configs/exp.yaml",
    #


def test_fednh():
    fednh, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fednh.yaml")
    # fednh, log = _test_algo("./tests/configs/exp.yaml",
    #                         "./tests/configs/alg/fednh.yaml", oncpu=False)


def test_fednova():
    fednova, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fednova.yaml")
    # fednova, log = _test_algo("./tests/configs/exp.yaml",
    #                           "./tests/configs/alg/fednova.yaml", oncpu=False)


def test_fedopt():
    fedopt, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedadam.yaml")
    fedopt, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedadagrad.yaml")
    fedopt, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedyogi.yaml")
    # fedopt, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/fedadam.yaml", oncpu=False)
    # fedopt, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/fedadagrad.yaml", oncpu=False)
    # fedopt, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/fedyogi.yaml", oncpu=False)


def test_fedproto():
    fedproto, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedproto.yaml")
    # fedproto, log = _test_algo("./tests/configs/exp.yaml",
    #                            "./tests/configs/alg/fedproto.yaml", oncpu=False)


def test_fedprox():
    fedprox, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedprox.yaml")
    # fedprox, log = _test_algo("./tests/configs/exp.yaml",
    #                           "./tests/configs/alg/fedprox.yaml", oncpu=False)


def test_fedper():
    fedper, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedper.yaml")
    # fedper, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/fedper.yaml", oncpu=False)


def test_fedrep():
    fedrep, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedrep.yaml")
    # fedrep, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/fedrep.yaml", oncpu=False)


def test_fedrod():
    fedrod, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedrod.yaml")
    # fedrod, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/fedrod.yaml", oncpu=False)


def test_fedrs():
    fedrs, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedrs.yaml")
    # fedrs, log = _test_algo("./tests/configs/exp.yaml",
    #                         "./tests/configs/alg/fedrs.yaml", oncpu=False)


def test_fedsam():
    fedsam, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedsam.yaml")
    # fedsam, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/fedsam.yaml", oncpu=False)


def test_fedsgd():
    fedsgd, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedsgd.yaml")
    # fedsgd, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/fedsgd.yaml", oncpu=False)


def test_gear():
    gear, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/gear.yaml")
    # gear, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/gear.yaml", oncpu=False)


def test_kafe():
    kafe, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/kafe.yaml")


def test_lgfedavg():
    lgfedavg, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/lg_fedavg.yaml")
    # lgfedavg, log = _test_algo("./tests/configs/exp.yaml",
    #                            "./tests/configs/alg/lg_fedavg.yaml", oncpu=False)


def test_moon():
    moon, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/moon.yaml")
    # moon, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/moon.yaml", oncpu=False)


def test_per_fedavg():
    per_fedavg, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/per_fedavg.yaml")
    # per_fedavg, log = _test_algo("./tests/configs/exp.yaml",
    #                              "./tests/configs/alg/per_fedavg.yaml", oncpu=False)
    per_fedavg, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/per_fedavg2.yaml")
    # per_fedavg, log = _test_algo("./tests/configs/exp.yaml",
    #                              "./tests/configs/alg/per_fedavg2.yaml", oncpu=False)


def test_pfedme():
    pfedme, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/pfedme.yaml")
    # pfedme, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/pfedme.yaml", oncpu=False)


def test_scaffold():
    scaffold, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/scaffold.yaml")
    # scaffold, log = _test_algo("./tests/configs/exp.yaml",
    #                            "./tests/configs/alg/scaffold.yaml", oncpu=False)


def test_superfed():
    superfed, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/superfed.yaml", 3)
    # superfed, log = _test_algo("./tests/configs/exp.yaml",
    #                            "./tests/configs/alg/superfed.yaml", 3, oncpu=False)


if __name__ == "__main__":
    # test_centralized_fl()
    # test_apfl()
    # test_ccvr()
    # test_ditto()
    # test_fat()
    test_fedamp()
    # test_fedavg()
    # test_fedavgm()
    # test_fedaws()
    # test_fedbabu()
    # test_fedbn()
    # test_feddyn()
    # test_fedexp()
    # test_fedhp()
    # test_fedlc()
    # test_fednh()
    # test_fednova()
    # test_fedopt()
    # test_fedper()
    # test_fedprox()
    # test_fedproto()
    # test_fedrep()
    # test_fedrod()
    # test_fedrs()
    # test_fedsam()
    # test_fedsgd()
    # test_gear()
    # test_kafe()
    # test_lgfedavg()
    # test_moon()
    # test_per_fedavg()
    # test_pfedme()
    # test_scaffold()
    # test_superfed()



from __future__ import annotations

import sys
import urllib
import warnings

import pytest
import torch
from torchvision.transforms import (Compose, Normalize, RandomCrop,
                                    RandomHorizontalFlip, ToTensor)

sys.path.append(".")
sys.path.append("..")

from fluke.data.datasets import Datasets  # NOQA
from fluke.data.support import CINIC10, MNISTM  # NOQA


# ### MNIST
def test_mnist():

    with pytest.raises(ValueError):
        mnist = Datasets.get("monist")

    mnist = Datasets.get("mnist", path="./data")
    assert mnist.train[0].shape == torch.Size([60000, 28, 28])
    assert mnist.train[1].shape == torch.Size([60000])
    assert mnist.test[0].shape == torch.Size([10000, 28, 28])
    assert mnist.test[1].shape == torch.Size([10000])
    assert mnist.num_classes == len(
        set(mnist.train[1].unique().tolist() + mnist.test[1].unique().tolist()))
    assert mnist.num_classes == 10

    mnist = Datasets.MNIST("./data")
    assert mnist.train[0].shape == torch.Size([60000, 28, 28])
    assert mnist.train[1].shape == torch.Size([60000])
    assert mnist.test[0].shape == torch.Size([10000, 28, 28])
    assert mnist.test[1].shape == torch.Size([10000])
    assert mnist.num_classes == len(
        set(mnist.train[1].unique().tolist() + mnist.test[1].unique().tolist()))
    assert mnist.num_classes == 10


# ### MNIST 4D
def test_mnist4d():
    mnist4d = Datasets.MNIST("./data", channel_dim=True)
    assert mnist4d.train[0].shape == torch.Size([60000, 1, 28, 28])
    assert mnist4d.train[1].shape == torch.Size([60000])
    assert mnist4d.test[0].shape == torch.Size([10000, 1, 28, 28])
    assert mnist4d.test[1].shape == torch.Size([10000])
    assert mnist4d.num_classes == len(
        set(mnist4d.train[1].unique().tolist() + mnist4d.test[1].unique().tolist()))
    assert mnist4d.num_classes == 10


# ### EMNIST
def test_emnist():
    emnist = Datasets.EMNIST("./data")
    assert emnist.train[0].shape == torch.Size([112800, 28, 28])
    assert emnist.train[1].shape == torch.Size([112800])
    assert emnist.test[0].shape == torch.Size([18800, 28, 28])
    assert emnist.test[1].shape == torch.Size([18800])
    assert emnist.num_classes == len(
        set(emnist.train[1].unique().tolist() + emnist.test[1].unique().tolist()))
    assert emnist.num_classes == 47


# ### SVHN
def test_svhn():
    try:
        svhn = Datasets.SVHN("./data")
    except urllib.error.URLError:
        warnings.warn("Skipped SVHN due to time out - Unknown issue!")
        assert True  # this is a "fix" to a non sensical issue on github action
        return
    assert svhn.train[0].shape == torch.Size([73257, 3, 32, 32])
    assert svhn.train[1].shape == torch.Size([73257])
    assert svhn.test[0].shape == torch.Size([26032, 3, 32, 32])
    assert svhn.test[1].shape == torch.Size([26032])
    assert svhn.num_classes == len(
        set(svhn.train[1].unique().tolist() + svhn.test[1].unique().tolist()))
    assert svhn.num_classes == 10


# ### CIFAR10
def test_cifar10():
    transform = Compose([
        RandomCrop(32, padding=4),
        RandomHorizontalFlip(),
        ToTensor(),
        Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
    ])
    cifar10 = Datasets.CIFAR10("./data", transforms=transform)
    assert cifar10.train[0].shape == torch.Size([50000, 3, 32, 32])
    assert cifar10.train[1].shape == torch.Size([50000])
    assert cifar10.test[0].shape == torch.Size([10000, 3, 32, 32])
    assert cifar10.test[1].shape == torch.Size([10000])
    assert cifar10.num_classes == len(
        set(cifar10.train[1].unique().tolist() + cifar10.test[1].unique().tolist()))
    assert cifar10.num_classes == 10

    cifar10 = Datasets.CIFAR10("./data")
    assert cifar10.train[0].shape == torch.Size([50000, 3, 32, 32])
    assert cifar10.train[1].shape == torch.Size([50000])
    assert cifar10.test[0].shape == torch.Size([10000, 3, 32, 32])
    assert cifar10.test[1].shape == torch.Size([10000])
    assert cifar10.num_classes == len(
        set(cifar10.train[1].unique().tolist() + cifar10.test[1].unique().tolist()))
    assert cifar10.num_classes == 10


# ### CIFAR100
def test_cifar100():
    cifar100 = Datasets.CIFAR100("./data")
    assert cifar100.train[0].shape == torch.Size([50000, 3, 32, 32])
    assert cifar100.train[1].shape == torch.Size([50000])
    assert cifar100.test[0].shape == torch.Size([10000, 3, 32, 32])
    assert cifar100.test[1].shape == torch.Size([10000])
    assert cifar100.num_classes == len(
        set(cifar100.train[1].unique().tolist() + cifar100.test[1].unique().tolist()))
    assert cifar100.num_classes == 100


# ### MNIST-M
def test_mnistm():

    dmnistm = MNISTM("./data", transform=ToTensor())
    img, label = dmnistm[0]
    assert img.shape == torch.Size([3, 28, 28])
    assert isinstance(label, int)
    assert dmnistm[0][0].shape == torch.Size([3, 28, 28])
    assert len(dmnistm) == 60000

    mnistm = Datasets.MNISTM("./data")
    assert mnistm.train[0].shape == torch.Size([60000, 3, 28, 28])
    assert mnistm.train[1].shape == torch.Size([60000])
    assert mnistm.test[0].shape == torch.Size([10000, 3, 28, 28])
    assert mnistm.test[1].shape == torch.Size([10000])
    assert mnistm.num_classes == len(
        set(mnistm.train[1].unique().tolist() + mnistm.test[1].unique().tolist()))
    assert mnistm.num_classes == 10


# ### Tiny Imagenet
def test_tinyimagenet():
    tiny_imagenet = Datasets.TINY_IMAGENET("./data")
    assert tiny_imagenet.train[0].shape == torch.Size([100000, 3, 64, 64])
    assert tiny_imagenet.train[1].shape == torch.Size([100000])
    assert tiny_imagenet.test[0].shape == torch.Size([10000, 3, 64, 64])
    assert tiny_imagenet.test[1].shape == torch.Size([10000])
    assert tiny_imagenet.num_classes == len(
        set(tiny_imagenet.train[1].unique().tolist() + tiny_imagenet.test[1].unique().tolist()))
    assert tiny_imagenet.num_classes == 200


# ### FEMNIST
def test_femnist():
    try:
        femnist = Datasets.FEMNIST("./data")
    except AssertionError:
        return
    assert len(femnist.clients_tr) == 3597  # Total number of clients
    assert len(femnist.clients_te) == 3597  # Total number of clients
    assert femnist.clients_tr[0].tensors[0].shape[1:] == torch.Size([1, 28, 28])  # image shape
    # number of labels matches number of images in each client
    assert sum([femnist.clients_tr[i].tensors[1].shape[0] == femnist.clients_tr[i].tensors[0].shape[0]
                for i in range(len(femnist.clients_tr))]) == len(femnist.clients_tr)

    assert femnist.clients_te[0].tensors[0].shape[1:] == torch.Size([1, 28, 28])  # image shape
    # number of labels matches number of images in each client
    assert sum([femnist.clients_te[i].tensors[1].shape[0] == femnist.clients_te[i].tensors[0].shape[0]
                for i in range(len(femnist.clients_te))]) == len(femnist.clients_te)

    lbl_train = set.union(*[set(femnist.clients_tr[i].tensors[1].numpy())
                          for i in range(len(femnist.clients_tr))])
    lbl_test = set.union(*[set(femnist.clients_te[i].tensors[1].numpy())
                         for i in range(len(femnist.clients_te))])
    assert len(lbl_train | lbl_test) == 62  # Total number of classes


def test_femnist_dig():
    try:
        femnist_dig = Datasets.FEMNIST("./data", filter="digits")
    except AssertionError:
        return
    assert len(femnist_dig.clients_tr) == 3597  # Total number of clients
    assert len(femnist_dig.clients_te) == 3597  # Total number of clients
    assert femnist_dig.clients_tr[0].tensors[0].shape[1:] == torch.Size([1, 28, 28])  # image shape
    # number of labels matches number of images in each client
    assert sum([femnist_dig.clients_tr[i].tensors[1].shape[0] == femnist_dig.clients_tr[i].tensors[0].shape[0]
                for i in range(len(femnist_dig.clients_tr))]) == len(femnist_dig.clients_tr)

    assert femnist_dig.clients_te[0].tensors[0].shape[1:] == torch.Size([1, 28, 28])  # image shape
    # number of labels matches number of images in each client
    assert sum([femnist_dig.clients_te[i].tensors[1].shape[0] == femnist_dig.clients_te[i].tensors[0].shape[0]
                for i in range(len(femnist_dig.clients_te))]) == len(femnist_dig.clients_te)

    lbl_train = set.union(*[set(femnist_dig.clients_tr[i].tensors[1].numpy())
                          for i in range(len(femnist_dig.clients_tr))])
    lbl_test = set.union(*[set(femnist_dig.clients_te[i].tensors[1].numpy())
                         for i in range(len(femnist_dig.clients_te))])
    assert len(lbl_train | lbl_test) == 10


def test_femnist_upp():
    try:
        femnist_u = Datasets.FEMNIST("./data", filter="uppercase")
    except AssertionError:
        return
    assert len(femnist_u.clients_tr) == 3597  # Total number of clients
    assert len(femnist_u.clients_te) == 3597  # Total number of clients
    assert femnist_u.clients_tr[0].tensors[0].shape[1:] == torch.Size([1, 28, 28])  # image shape
    # number of labels matches number of images in each client
    assert sum([femnist_u.clients_tr[i].tensors[1].shape[0] == femnist_u.clients_tr[i].tensors[0].shape[0]
                for i in range(len(femnist_u.clients_tr))]) == len(femnist_u.clients_tr)

    assert femnist_u.clients_te[0].tensors[0].shape[1:] == torch.Size([1, 28, 28])  # image shape
    # number of labels matches number of images in each client
    assert sum([femnist_u.clients_te[i].tensors[1].shape[0] == femnist_u.clients_te[i].tensors[0].shape[0]
                for i in range(len(femnist_u.clients_te))]) == len(femnist_u.clients_te)

    lbl_train = set.union(*[set(femnist_u.clients_tr[i].tensors[1].numpy())
                          for i in range(len(femnist_u.clients_tr))])
    lbl_test = set.union(*[set(femnist_u.clients_te[i].tensors[1].numpy())
                         for i in range(len(femnist_u.clients_te))])
    assert len(lbl_train | lbl_test) == 26


def test_femnist_low():
    try:
        femnist_l = Datasets.FEMNIST("./data", filter="lowercase")
    except AssertionError:
        return
    assert len(femnist_l.clients_tr) == 3597  # Total number of clients
    assert len(femnist_l.clients_te) == 3597  # Total number of clients
    assert femnist_l.clients_tr[0].tensors[0].shape[1:] == torch.Size([1, 28, 28])  # image shape
    # number of labels matches number of images in each client
    assert sum([femnist_l.clients_tr[i].tensors[1].shape[0] == femnist_l.clients_tr[i].tensors[0].shape[0]
                for i in range(len(femnist_l.clients_tr))]) == len(femnist_l.clients_tr)

    assert femnist_l.clients_te[0].tensors[0].shape[1:] == torch.Size([1, 28, 28])  # image shape
    # number of labels matches number of images in each client
    assert sum([femnist_l.clients_te[i].tensors[1].shape[0] == femnist_l.clients_te[i].tensors[0].shape[0]
                for i in range(len(femnist_l.clients_te))]) == len(femnist_l.clients_te)

    lbl_train = set.union(*[set(femnist_l.clients_tr[i].tensors[1].numpy())
                          for i in range(len(femnist_l.clients_tr))])
    lbl_test = set.union(*[set(femnist_l.clients_te[i].tensors[1].numpy())
                         for i in range(len(femnist_l.clients_te))])
    assert len(lbl_train | lbl_test) == 26


# ### Shakespeare
def test_shakespeare():
    try:
        shake = Datasets.SHAKESPEARE("./data")
    except AssertionError:
        return

    assert len(shake.clients_tr) == 660
    assert len(shake.clients_te) == 660
    assert shake.clients_tr[0].tensors[0].shape[1:] == torch.Size([80])  # Shakespeare text
    assert shake.clients_te[0].tensors[0].shape[1:] == torch.Size([80])  # Shakespeare text

    assert sum([shake.clients_tr[i].tensors[1].shape[0] == shake.clients_tr[i].tensors[0].shape[0]
                for i in range(len(shake.clients_tr))]) == len(shake.clients_tr)
    assert sum([shake.clients_te[i].tensors[1].shape[0] == shake.clients_te[i].tensors[0].shape[0]
                for i in range(len(shake.clients_te))]) == len(shake.clients_te)


# ### Fashion MNIST
def test_fashion_mnist():
    fashion = Datasets.FASHION_MNIST("./data")
    assert fashion.train[0].shape == torch.Size([60000, 28, 28])
    assert fashion.train[1].shape == torch.Size([60000])
    assert fashion.test[0].shape == torch.Size([10000, 28, 28])
    assert fashion.test[1].shape == torch.Size([10000])
    assert fashion.num_classes == len(set(fashion.train[1].unique().tolist() +
                                          fashion.test[1].unique().tolist()))
    assert fashion.num_classes == 10


# ### CINIC10
def test_cinic10():
    dcinic = CINIC10("./data")
    img, label = dcinic[0]
    assert img.shape == torch.Size([3, 32, 32])
    assert label.shape == torch.Size([])
    assert len(dcinic) == 90000

    cinic = Datasets.CINIC10("./data")
    assert cinic.train[0].shape == torch.Size([90000, 3, 32, 32])
    assert cinic.train[1].shape == torch.Size([90000])
    assert cinic.test[0].shape == torch.Size([90000, 3, 32, 32])
    assert cinic.test[1].shape == torch.Size([90000])
    assert cinic.num_classes == len(
        set(cinic.train[1].unique().tolist() + cinic.test[1].unique().tolist()))
    assert cinic.num_classes == 10


# ### CINIC10
def test_fcube():
    data_container = Datasets.FCUBE()
    ctr, cte, ste = data_container.clients_tr, data_container.clients_te, data_container.server_data
    assert len(ctr) == 4
    assert len(cte) == 4
    assert ste is not None
    assert ste.size == 100
    assert sum([cte[i].size + ctr[i].size for i in range(4)]) == 900
    assert ctr[0].num_labels == 2


if __name__ == "__main__":
    # test_mnist()
    # test_mnist4d()
    # test_emnist()
    # test_svhn()
    # test_cifar10()
    # test_cifar100()
    # test_mnistm()
    # test_tinyimagenet()
    # test_femnist()
    # test_femnist_dig()
    # test_femnist_upp()
    # test_femnist_low()
    # test_shakespeare()
    # test_fashion_mnist()
    # test_cinic10()
    test_fcube()

    # 98% coverate on datasets.py
    # 88% coverage on support.py



"""Implementation of the FAT [FAT20]_ algorithm.

References:
    .. [FAT20] Giulio Zizzo, Ambrish Rawat, Mathieu Sinn, Beat Buesser.
       FAT: Federated Adversarial Training. In SpicyFL@NeurIPS (2020).
       URL: https://arxiv.org/abs/2012.01791

"""
import sys
from typing import Any

import numpy as np
import torch
import torch.nn.functional as F
from torch.nn import Module

sys.path.append(".")
sys.path.append("..")

from ..client import Client  # NOQA
from ..data import FastDataLoader  # NOQA
from ..utils import OptimizerConfigurator, clear_cuda_cache  # NOQA
from . import CentralizedFL  # NOQA

__all__ = [
    "FATClient",
    "FAT"
]


class FATClient(Client):

    def __init__(self,
                 index: int,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: Module,  # not used
                 local_epochs: int,
                 fine_tuning_epochs: int = 0,
                 clipping: float = 0,
                 eps: float = 0.1,
                 alpha: float = 2.0 / 255,
                 adv_iters: int = 10,
                 k_prop: float = 1.0,
                 **kwargs: dict[str, Any]):
        assert 0.0 <= k_prop <= 1.0, "k_prop must be in [0, 1]"
        self.sample_per_class = torch.zeros(train_set.num_labels)
        uniq_val, uniq_count = np.unique(train_set.tensors[1], return_counts=True)
        for i, c in enumerate(uniq_val.tolist()):
            self.sample_per_class[c] = uniq_count[i]

        super().__init__(index=index, train_set=train_set, test_set=test_set,
                         loss_fn=loss_fn,
                         optimizer_cfg=optimizer_cfg,
                         local_epochs=local_epochs,
                         fine_tuning_epochs=fine_tuning_epochs,
                         clipping=clipping,
                         **kwargs)
        self.hyper_params.update(eps=eps, alpha=alpha, adv_iters=adv_iters, k_prop=k_prop)

    def generate_adversarial(self,
                             model: Module,
                             inputs: torch.Tensor,
                             targets: torch.Tensor,
                             eps: float = 0.1,
                             alpha: float = 2.0 / 255,
                             iters: int = 10) -> torch.Tensor:
        """
        Generates adversarial examples using Projected Gradient Descent (PGD).

        Args:
            model (nn.Module): The neural network model.
            inputs (Tensor): Input samples.
            targets (Tensor): True labels for the inputs.
            alpha (float): Step size for each iteration.
            iters (int): Number of attack iterations.

        Returns:
            Tensor: Adversarial examples.
        """
        adv_inputs = inputs.clone().detach().requires_grad_(True)
        for _ in range(iters):
            outputs = model(adv_inputs)
            loss = F.cross_entropy(outputs, targets)
            model.zero_grad()
            loss.backward()
            perturbation = alpha * adv_inputs.grad.sign()
            adv_inputs = adv_inputs + perturbation
            adv_inputs = torch.min(torch.max(adv_inputs, inputs - eps), inputs + eps)
            adv_inputs = torch.clamp(adv_inputs, 0, 1).detach().requires_grad_(True)
        return adv_inputs.requires_grad_(False)

    def fit(self, override_local_epochs: int = 0) -> float:
        epochs: int = (override_local_epochs if override_local_epochs > 0
                       else self.hyper_params.local_epochs)

        self.model.train()
        self.model.to(self.device)

        if self.optimizer is None:
            self.optimizer, self.scheduler = self._optimizer_cfg(self.model)

        running_loss = 0.0
        k = int(self.hyper_params.k_prop * self.train_set.batch_size)
        for _ in range(epochs):
            for _, (X, y) in enumerate(self.train_set):
                X, y = X.to(self.device), y.to(self.device)
                X_adv = self.generate_adversarial(self.model, X[:k], y[:k])
                X = torch.cat((X, X_adv), dim=0)
                y = torch.cat((y, y[:k]), dim=0)
                self.optimizer.zero_grad()
                y_hat = self.model(X)
                loss = self.hyper_params.loss_fn(y_hat, y)
                loss.backward()
                self._clip_grads(self.model)
                self.optimizer.step()
                running_loss += loss.item()
            self.scheduler.step()

        running_loss /= (epochs * len(self.train_set))
        self.model.cpu()
        clear_cuda_cache()
        return running_loss


class FAT(CentralizedFL):

    def get_client_class(self):
        return FATClient



"""Implementation of the Federated Averaging with Spreadout [FedAwS20]_ algorithm.

References:
    .. [FedAwS20] Felix X. Yu, Ankit Singh Rawat, Aditya Krishna Menon, Sanjiv Kumar.
       Federated Learning with Only Positive Labels.
       In ICML (2020). URL: https://proceedings.mlr.press/v119/yu20f/yu20f.pdf
"""
import sys
from typing import Any, Iterable, Literal

import torch
from torch import nn
from torch.nn import functional as F

sys.path.append(".")
sys.path.append("..")

from ..client import Client  # NOQA
from ..data import FastDataLoader  # NOQA
from ..server import Server  # NOQA
from . import CentralizedFL  # NOQA

__all__ = [
    "SpreadModel",
    "SpreadLoss",
    "FedAwSServer",
    "FedAwS"
]


class SpreadModel(nn.Module):
    def __init__(self, weights: torch.Tensor):
        super().__init__()
        self.weights = nn.Parameter(weights)

    def forward(self):
        return F.normalize(self.weights, dim=1)


class SpreadLoss(nn.Module):

    def __init__(self, margin: float, reduction: Literal["mean", "sum"] = "mean"):
        super().__init__()
        self.margin = margin
        self.reduction = reduction

    def forward(self, weights: torch.Tensor):
        ws_norm = F.normalize(weights, dim=1)
        cos_dis = 0.5 * (1.0 - torch.mm(ws_norm, ws_norm.transpose(0, 1)))

        d_mat = torch.diag(torch.ones(weights.shape[0]))
        d_mat = d_mat.to(weights.device)

        cos_dis = cos_dis * (1.0 - d_mat)

        indx = ((self.margin - cos_dis) > 0.0).float()
        loss = (((self.margin - cos_dis) * indx) ** 2)

        if self.reduction == "mean":
            return loss.mean()
        elif self.reduction == "sum":
            return loss.sum()

        return loss


class FedAwSServer(Server):

    def __init__(self,
                 model: torch.nn.Module,
                 test_set: FastDataLoader,
                 clients: Iterable[Client],
                 weighted: bool = False,
                 aws_lr: float = 0.1,
                 aws_steps: int = 100,
                 margin: float = 0.5,
                 last_layer_name: str = "classifier",
                 **kwargs: dict[str, Any]):
        super().__init__(model=model, test_set=test_set, clients=clients, weighted=weighted)
        assert (last_layer_name + ".weight") in model.state_dict().keys(), \
            f"Invalid last_layer_name: {last_layer_name}. Make sure that the last layer \
                is named as {last_layer_name}"
        self.hyper_params.update(aws_lr=aws_lr,
                                 margin=margin,
                                 aws_steps=aws_steps,
                                 last_layer_name=last_layer_name + ".weight")

    def _compute_spreadout(self):
        ws = self.model.state_dict()[self.hyper_params.last_layer_name].data
        spread_model = SpreadModel(ws)

        optimizer = torch.optim.SGD(
            spread_model.parameters(),
            lr=self.hyper_params.aws_lr,
            momentum=0.9
        )
        criterion = SpreadLoss(margin=self.hyper_params.margin)

        for _ in range(self.hyper_params.aws_steps):
            optimizer.zero_grad()
            loss = criterion(spread_model.forward())
            loss.backward()
            optimizer.step()

        self.model.load_state_dict(
            {self.hyper_params.last_layer_name: spread_model.weights.data},
            strict=False
        )

    def aggregate(self, eligible: Iterable[Client], client_models: Iterable[nn.Module]) -> None:
        super().aggregate(eligible, client_models)
        self._compute_spreadout()


class FedAwS(CentralizedFL):

    def get_server_class(self) -> Server:
        return FedAwSServer



"""Implementation of the FedExP [FedExP23]_ algorithm.

References:
    .. [FedExP23] Divyansh Jhunjhunwala, Shiqiang Wang, and Gauri Joshi.
       FedExP: Speeding Up Federated Averaging via Extrapolation. In ICLR (2023).
       URL: https://arxiv.org/abs/2301.09604
"""
import sys
from typing import Iterable

import torch
from torch.nn import Module

sys.path.append(".")
sys.path.append("..")

from ..algorithms import CentralizedFL  # NOQA
from ..client import Client  # NOQA
from ..server import Server  # NOQA
from ..utils.model import flatten_parameters  # NOQA

__all__ = [
    "FedExPServer",
    "FedExP"
]


class FedExPServer(Server):

    @torch.no_grad()
    def aggregate(self, eligible: Iterable[Client], client_models: Iterable[Module]) -> None:
        W = flatten_parameters(self.model)
        client_models = list(client_models)
        Wi = [flatten_parameters(client_model) for client_model in client_models]
        eta = self._compute_eta(W, Wi)

        self.hyper_params.update(lr=eta)
        super().aggregate(eligible, client_models)

    def _compute_eta(self, W: torch.Tensor, Wi: list[torch.Tensor], eps: float = 1e-4) -> float:
        Delta_bar = torch.mean(W - torch.stack(Wi), dim=0)
        sum_norm_Delta_i = torch.sum(torch.norm(W - torch.stack(Wi))**2, dim=0)
        norm_Delta_bar = torch.norm(Delta_bar)**2
        eta = torch.max(sum_norm_Delta_i / (2 * len(Wi) * (norm_Delta_bar + eps)),
                        torch.FloatTensor([1.0]))
        return eta


class FedExP(CentralizedFL):

    def get_server_class(self) -> Server:
        return FedExPServer



"""Implementation of the [pFedMe20]_ algorithm.

References:
    .. [pFedMe20] Canh T. Dinh, Nguyen H. Tran, and Tuan Dung Nguyen. Personalized Federated
       Learning with Moreau Envelopes. In NeurIPS (2020). URL: https://arxiv.org/abs/2006.08848
"""
from ..utils.model import safe_load_state_dict
import sys
from copy import deepcopy
from typing import Any, Iterable, Optional

import torch
from torch.nn import Module
from torch.optim import Optimizer

sys.path.append(".")
sys.path.append("..")

from ..algorithms import CentralizedFL  # NOQA
from ..client import Client  # NOQA
from ..data import FastDataLoader  # NOQA
from ..server import Server  # NOQA
from ..utils import OptimizerConfigurator  # NOQA
from ..utils.model import aggregate_models  # NOQA

__all__ = [
    "PFedMeOptimizer",
    "PFedMeClient",
    "PFedMeServer",
    "PFedMe"
]


class PFedMeOptimizer(Optimizer):
    def __init__(self, params, lr=0.01, lamda=0.1, mu=0.001):
        defaults = dict(lr=lr, lamda=lamda, mu=mu)
        super(PFedMeOptimizer, self).__init__(params, defaults)

    def step(self,
             local_parameters: list[torch.nn.Parameter],
             closure: callable = None) -> Optional[float]:
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure
        for group in self.param_groups:
            for param_p, param_l in zip(group["params"], local_parameters):
                param_p.data = param_p.data - group["lr"] * (
                    param_p.grad.data
                    + group["lamda"] * (param_p.data - param_l.data)
                    + group["mu"] * param_p.data
                )
        return loss


class PFedMeClient(Client):
    def __init__(self,
                 index: int,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: torch.nn.Module,
                 local_epochs: int,
                 k: int,
                 fine_tuning_epochs: int = 0,
                 clipping: float = 0,
                 **kwargs: dict[str, Any]):

        super().__init__(index=index, train_set=train_set, test_set=test_set,
                         optimizer_cfg=optimizer_cfg, loss_fn=loss_fn, local_epochs=local_epochs,
                         fine_tuning_epochs=fine_tuning_epochs, clipping=clipping, **kwargs)
        self.hyper_params.update(k=k)
        self.internal_model: Module = None
        self._attr_to_cache.append("internal_model")

    def receive_model(self) -> None:
        model = self.channel.receive(self, self.server, msg_type="model").payload
        if self.model is None:
            self.model = model
            self.internal_model = deepcopy(model)
        else:
            safe_load_state_dict(self.model, model.state_dict())

    def fit(self, override_local_epochs: int = 0) -> float:
        epochs: int = (override_local_epochs if override_local_epochs > 0
                       else self.hyper_params.local_epochs)
        self.internal_model.to(self.device)
        self.model.train()
        self.model.to(self.device)

        if self.optimizer is None:
            self.optimizer, self.scheduler = self._optimizer_cfg(self.model)

        lamda = self.optimizer.defaults["lamda"]
        running_loss = 0.0
        for _ in range(epochs):
            for _, (X, y) in enumerate(self.train_set):
                X, y = X.to(self.device), y.to(self.device)
                for _ in range(self.hyper_params.k):
                    self.optimizer.zero_grad()
                    y_hat = self.model(X)
                    loss = self.hyper_params.loss_fn(y_hat, y)
                    loss.backward()
                    self._clip_grads(self.model)
                    self.optimizer.step(self.model.parameters())

                lr = self.optimizer.param_groups[0]["lr"]
                params = zip(self.model.parameters(), self.internal_model.parameters())
                for param_p, param_l in params:
                    param_l.data = param_l.data - lamda * lr * (param_l.data - param_p.data)
                running_loss += loss.item()
            self.scheduler.step()
        running_loss /= (epochs * len(self.train_set))
        self.internal_model.load_state_dict(self.model.state_dict())
        self.internal_model.cpu()
        self.model.cpu()
        return running_loss


class PFedMeServer(Server):
    def __init__(self,
                 model: Module,
                 test_set: FastDataLoader,
                 clients: Iterable[Client],
                 weighted: bool = False,
                 beta: float = 0.5,
                 **kwargs: dict[str, Any]):
        super().__init__(model=model, test_set=test_set, clients=clients, weighted=weighted)
        self.hyper_params.update(beta=beta)

    @torch.no_grad()
    def aggregate(self, eligible: Iterable[Client], client_models: Iterable[Module]) -> None:
        weights = self._get_client_weights(eligible)
        agg_model_sd = aggregate_models(self.model,
                                        client_models,
                                        weights,
                                        self.hyper_params.lr,
                                        inplace=False).state_dict()

        for key, param in self.model.named_parameters():
            param.data = (1 - self.hyper_params.beta) * param.data
            param.data += self.hyper_params.beta * agg_model_sd[key]


class PFedMe(CentralizedFL):

    def can_override_optimizer(self) -> bool:
        return False

    def get_optimizer_class(self) -> torch.optim.Optimizer:
        return PFedMeOptimizer

    def get_client_class(self) -> Client:
        return PFedMeClient

    def get_server_class(self) -> Server:
        return PFedMeServer



from __future__ import annotations

import os
import shutil
import sys

import torch
from rich.live import Live
from rich.progress import Progress

sys.path.append(".")
sys.path.append("..")

from fluke import DDict, FlukeENV, ObserverSubject, FlukeCache  # NOQA
from fluke.utils import Configuration  # NOQA


def test_env():
    cfg = Configuration("tests/configs/exp.yaml", "tests/configs/alg/fedavg.yaml")
    env = FlukeENV()
    env.set_seed(117)
    assert env.get_seed() == 117
    env.set_device("cpu")
    assert env.get_device() == torch.device("cpu")
    env.set_device("cuda")
    assert env.get_device() == torch.device("cuda")
    env.set_device("mps")
    assert env.get_device() == torch.device("mps")
    env.set_device("auto")
    env.set_eval_cfg(DDict(pre_fit=True))
    assert env.get_eval_cfg().pre_fit
    env.set_save_options(path="temp", save_every=10, global_only=True)
    assert env.get_save_options() == ("temp", 10, True)
    if torch.cuda.is_available():
        assert env.get_device() == torch.device("cuda")
    elif torch.backends.mps.is_available():
        assert env.get_device() == torch.device("mps")
    else:
        assert env.get_device() == torch.device("cpu")
    live = env.get_live_renderer()
    assert live is not None and isinstance(live, Live)
    progress_fl = env.get_progress_bar("FL")
    progress_s = env.get_progress_bar("clients")
    progress_c = env.get_progress_bar("server")
    assert progress_fl is not None and isinstance(progress_fl, Progress)
    assert progress_s is not None and isinstance(progress_s, Progress)
    assert progress_c is not None and isinstance(progress_c, Progress)
    assert progress_fl != progress_s
    assert progress_fl != progress_c
    assert progress_s != progress_c

    env.configure(cfg)
    assert env.get_seed() == 42

    env.force_close()

    env.open_cache("test_env")
    assert env.get_cache() is None

    shutil.rmtree("tmp/tmp", ignore_errors=True)

    env.set_inmemory(False)
    env.open_cache("test_env")
    assert env.get_cache() is not None
    assert os.path.exists(env.get_cache().cache_dir)

    cache = env.get_cache()
    print(list(cache._cache.iterkeys()))
    assert cache.occupied == 0
    ref = cache.push("test1", "this is a test")
    assert isinstance(ref, FlukeCache._ObjectRef)
    assert cache.occupied == 1
    assert cache.get("test1") == "this is a test"
    assert cache.get("test2") is None
    assert cache.get("test2", "default") == "default"
    cache.push("test2", ref)
    assert cache.occupied == 1
    assert cache._cache[ref.id].refs == 2

    test1 = cache.pop("test1")
    assert test1 == "this is a test"
    assert cache.occupied == 1
    assert cache._cache[ref.id].refs == 1

    cache.cleanup()
    assert cache.occupied == 1
    cache.delete("test2")
    assert cache.occupied == 0
    cache.cleanup()
    assert cache.occupied == 0
    FlukeENV().close_cache()


def test_observer():
    subj = ObserverSubject()
    assert subj._observers == []
    subj.attach("test")
    assert subj._observers == ["test"]
    subj.detach("test")
    assert subj._observers == []

    subj.detach("test2")


def test_ddict():
    dd = DDict(**{
        "a": 1,
        "b": 2,
        "c": {
            "d": 3,
            "e": 4
        }
    })

    assert dd.a == 1
    assert dd.b == 2
    assert dd.c.d == 3
    assert dd.c.e == 4

    dd_nota = dd.exclude("a")
    assert "a" not in dd_nota
    assert dd_nota.b == 2

    dd = DDict({"a": 1, "b": 2}, c=3)
    assert dd.a == 1
    assert dd.b == 2
    assert dd.c == 3

    dd = DDict({"a": 1, "b": 2}, c=3)
    dd2 = DDict({"a": 1, "b": 2}, c=3)
    assert dd.match(dd2)
    assert dd2.match(dd, full=True)
    assert dd.diff(dd2) == {}
    d = DDict(a=1, b=2, c=3, e=DDict(a=1, b=2, c=3))
    e = DDict(a=1, b=3, c=4, e=DDict(a=1, b=1))
    assert d.diff(e) == {'b': 3, 'c': 4, 'e': {'b': 1}}
    assert e.diff(d) == {'b': 2, 'c': 3, 'e': {'b': 2, 'c': 3}}
    assert not d.match(e)


if __name__ == "__main__":
    test_env()
    test_observer()
    test_ddict()
    # coverage: 94% __init__.py



import sys

import pytest
import torch
from torch.nn import CrossEntropyLoss, Linear
from torch.optim import SGD

sys.path.append(".")
sys.path.append("..")

from fluke import DDict, FlukeENV  # NOQA
from fluke.client import Client, PFLClient  # NOQA
from fluke.data import FastDataLoader  # NOQA
from fluke.evaluation import ClassificationEval  # NOQA
from fluke.server import Server  # NOQA
from fluke.utils import OptimizerConfigurator  # NOQA


def test_client_cache():
    _test_client(False)


def test_client():
    _test_client(True)


class Model(Linear):
    def __init__(self):
        super().__init__(10, 2)

        # initialize weights to 0
        self.weight.data.fill_(0)
        self.bias.data.fill_(0)


def _test_client(inmemory):

    FlukeENV().set_inmemory(inmemory)
    if not inmemory:
        FlukeENV().open_cache("test_client")
    FlukeENV().set_eval_cfg(DDict(pre_fit=True, post_fit=True))

    # function that taken a 10-dimensional input returns a 0 if the
    # sum of the first 7 elements is less than 2.5
    def target_function(x):
        return 0 if x[:7].sum() < 2.5 else 1

    Xtr = torch.rand((100, 10))
    ytr = torch.tensor([target_function(x) for x in Xtr])
    Xte = torch.rand((100, 10))
    yte = torch.tensor([target_function(x) for x in Xte])
    train_set = FastDataLoader(
        Xtr, ytr,
        num_labels=2,
        batch_size=10,
        shuffle=True
    )

    test_set = FastDataLoader(
        Xte, yte,
        num_labels=2,
        batch_size=10,
        shuffle=True
    )

    client = Client(
        index=0,
        train_set=train_set,
        test_set=test_set,
        clipping=10,
        optimizer_cfg=OptimizerConfigurator(
            optimizer_cfg=DDict(name=SGD, lr=0.1),
            scheduler_cfg=DDict(step_size=1, gamma=0.1),
        ),
        loss_fn=CrossEntropyLoss(),
        local_epochs=10
    )

    assert client.index == 0
    assert client.train_set == train_set
    assert client.test_set == test_set
    assert client.n_examples == 100

    server = Server(
        model=Model(),
        test_set=None,
        clients=[client]
    )

    client.set_server(server)
    assert client.server == server
    assert client.channel == server.channel
    server.broadcast_model([client])

    evaluator = ClassificationEval(1, 2)
    FlukeENV().set_evaluator(evaluator)
    ev0 = client.evaluate(evaluator, client.test_set)
    client.local_update(1)
    ev1 = client.evaluate(evaluator, client.test_set)
    assert server.channel._buffer[server]
    assert not ev0
    assert ev1
    # if inmemory:
    assert isinstance(client.optimizer, SGD)
    assert client.optimizer.defaults["lr"] == 0.1
    assert isinstance(client.scheduler, torch.optim.lr_scheduler.StepLR)
    assert client.scheduler.step_size == 1

    assert str(client) == "Client[0](optim=OptCfg(SGD, lr=0.1, StepLR(step_size=1, gamma=0.1)), " + \
        "batch_size=10, loss_fn=CrossEntropyLoss(), local_epochs=10, fine_tuning_epochs=0, clipping=10)"

    client.test_set = None  # THIS IS NOT ALLOWED
    assert client.evaluate(evaluator, client.test_set) == {}

    server.broadcast_model([client])
    client.receive_model()
    # assert client.model.weight.data.sum() == 0
    # assert client.model.bias.data.sum() == 0

    client.send_model()

    m = server.channel.receive(server, client, "model")
    assert id(m) != id(client.model)
    assert m is not client.model
    assert torch.all(client.local_model.weight.data == client.model.weight.data)
    assert torch.all(client.local_model.bias.data == client.model.bias.data)

    server.broadcast_model([client])
    client.local_update(1)
    server.broadcast_model([client])
    client.finalize()
    assert client.model.weight.data.sum() == 0
    assert client.model.bias.data.sum() == 0

    assert str(client) == "Client[0](optim=OptCfg(SGD, lr=0.1, StepLR(step_size=1, gamma=0.1)), " + \
        "batch_size=10, loss_fn=CrossEntropyLoss(), local_epochs=10, fine_tuning_epochs=0, clipping=10)"
    assert str(client) == repr(client)

    client.save("tmp/client")
    client2 = Client(
        index=0,
        train_set=train_set,
        test_set=test_set,
        clipping=10,
        optimizer_cfg=OptimizerConfigurator(
            optimizer_cfg=DDict(name=SGD, lr=0.1),
            scheduler_cfg=DDict(step_size=1, gamma=0.1),
        ),
        loss_fn=CrossEntropyLoss(),
        local_epochs=10
    )
    client2.load("tmp/client", Model())
    assert torch.all(client2.model.weight.data == client.model.weight.data)

    if not inmemory:
        FlukeENV().close_cache()


def test_pflclient():

    class Model(Linear):
        def __init__(self):
            super().__init__(10, 2)

    # function that taken a 10-dimensional input returns a 0 if the
    # sum of the first 7 elements is less than 2.5
    def target_function(x):
        return 0 if x[:7].sum() < 2.5 else 1

    Xtr = torch.rand((100, 10))
    ytr = torch.tensor([target_function(x) for x in Xtr])
    train_set = FastDataLoader(
        Xtr, ytr,
        num_labels=2,
        batch_size=10,
        shuffle=True
    )

    client = PFLClient(
        index=0,
        model=Model(),
        train_set=train_set,
        test_set=train_set,
        optimizer_cfg=OptimizerConfigurator(
            optimizer_cfg=DDict(name=SGD, lr=0.1),
            scheduler_cfg=DDict(step_size=1, gamma=0.1)
        ),
        loss_fn=CrossEntropyLoss(),
        local_epochs=10
    )

    assert client.model is None
    assert client.personalized_model is not None, isinstance(client.personalized_model, Model)

    evaluator = ClassificationEval(1, 2)
    try:
        client.evaluate(evaluator, client.test_set)
    except Exception:
        pytest.fail("Unexpected exception!")
    client.test_set = None  # THIS IS NOT ALLOWED
    assert client.evaluate(evaluator, client.test_set) == {}

    server = Server(
        model=Model(),
        test_set=None,
        clients=[client]
    )

    client.set_server(server)
    client.pers_optimizer = SGD(client.personalized_model.parameters(), lr=0.1)
    client.pers_scheduler = torch.optim.lr_scheduler.StepLR(
        client.pers_optimizer, step_size=1, gamma=0.1)
    server.broadcast_model([client])
    client.local_update(1)
    client.save("tmp/pflclient")
    client.fit(1)

    client2 = PFLClient(
        index=0,
        model=Model(),
        train_set=train_set,
        test_set=train_set,
        optimizer_cfg=OptimizerConfigurator(
            optimizer_cfg=DDict(name=SGD, lr=0.1),
            scheduler_cfg=DDict(step_size=1, gamma=0.1)
        ),
        loss_fn=CrossEntropyLoss(),
        local_epochs=10
    )
    client2.load("tmp/pflclient", Model())
    assert torch.all(client2.personalized_model.weight.data ==
                     client.personalized_model.weight.data)
    assert torch.all(client2.model.weight.data ==
                     client.model.weight.data)


if __name__ == "__main__":
    test_client()
    # 100% coverage for client.py



"""Implementation of the FedDyn [FedDyn21]_ algorithm.

References:
    .. [FedDyn21] Durmus Alp Emre Acar, Yue Zhao, Ramon Matas, Matthew Mattina, Paul Whatmough,
       and Venkatesh Saligrama. Federated Learning with Dynamic Regularization.
       In ICLR (2021). URL: https://openreview.net/pdf?id=B7v4QMR6Z9w
"""
import sys
from copy import deepcopy
from typing import Any, Iterable

import numpy as np
import torch
from torch.nn import Module

sys.path.append(".")
sys.path.append("..")

from .. import FlukeENV  # NOQA
from ..client import Client  # NOQA
from ..comm import Message  # NOQA
from ..data import FastDataLoader  # NOQA
from ..server import Server  # NOQA
from ..utils import OptimizerConfigurator, clear_cuda_cache  # NOQA
from ..utils.model import aggregate_models, safe_load_state_dict  # NOQA
from . import CentralizedFL  # NOQA

__all__ = [
    "FedDynClient",
    "FedDynServer",
    "FedDyn"
]


def get_all_params_of(model: torch.Tensor, copy: bool = True) -> torch.Tensor:
    result = None
    for param in model.parameters():
        if not param.requires_grad:
            continue

        if result is None:
            result = param.clone().detach().reshape(-1) if copy else param.reshape(-1)
        else:
            result = torch.cat((result, param.clone().detach().reshape(-1)),
                               0) if copy else torch.cat((result, param.reshape(-1)), 0)
    return result


def load_all_params(device: torch.device, model: torch.nn.Module, params: torch.Tensor) -> None:
    dict_param = deepcopy(dict(model.named_parameters()))
    idx = 0
    for name, param in model.named_parameters():
        if not param.requires_grad:
            continue
        weights = param.data
        length = len(weights.reshape(-1))
        dict_param[name].data.copy_(
            params[idx:idx+length].clone().detach().reshape(weights.shape).to(device))
        idx += length

    model.load_state_dict(dict_param, strict=False)


class FedDynClient(Client):

    def __init__(self,
                 index: int,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: torch.nn.Module,
                 local_epochs: int,
                 alpha: float,
                 fine_tuning_epochs: int = 0,
                 clipping: float = 5,
                 **kwargs: dict[str, Any]):
        super().__init__(index=index, train_set=train_set, test_set=test_set,
                         optimizer_cfg=optimizer_cfg, loss_fn=loss_fn, local_epochs=local_epochs,
                         fine_tuning_epochs=fine_tuning_epochs, clipping=clipping, **kwargs)

        self.hyper_params.update(alpha=alpha)
        self.weight = None
        self.weight_decay = self._optimizer_cfg.optimizer_cfg[
            "weight_decay"] if "weight_decay" in self._optimizer_cfg.optimizer_cfg else 0
        self.prev_grads = None
        self._attr_to_cache.extend(["prev_grads", "weight"])

    def receive_model(self) -> None:
        model, cld_mdl = self.channel.receive(self, self.server, msg_type="model").payload
        if self.model is None:
            self.model = model
            self.prev_grads = torch.zeros_like(get_all_params_of(self.model), device=self.device)
        else:
            safe_load_state_dict(self.model, cld_mdl.state_dict())

    def _receive_weights(self) -> None:
        self.weight = self.channel.receive(self, self.server, msg_type="weight").payload

    def _send_weight(self) -> None:
        self.channel.send(Message(self.train_set.tensors[0].shape[0],
                                  "weight",
                                  self,
                                  inmemory=True), self.server)

    def send_model(self) -> None:
        self.channel.send(Message(self.model, "model", self, inmemory=True), self.server)
        self.channel.send(Message(self.prev_grads, "grads", self, inmemory=True), self.server)

    def fit(self, override_local_epochs: int = 0) -> float:
        epochs: int = (override_local_epochs if override_local_epochs > 0
                       else self.hyper_params.local_epochs)

        self.model.train()
        self.model.to(self.device)

        alpha_coef_adpt = self.hyper_params.alpha / self.weight
        server_params = get_all_params_of(self.model).to(self.device)

        for params in self.model.parameters():
            params.requires_grad = True

        if not self.optimizer:
            w_dec = alpha_coef_adpt + self.weight_decay
            self.optimizer, self.scheduler = self._optimizer_cfg(self.model,
                                                                 # this override the weight_decay
                                                                 # in the optimizer_cfg
                                                                 weight_decay=w_dec)
        running_loss = 0.0
        for _ in range(epochs):
            for _, (X, y) in enumerate(self.train_set):
                X, y = X.to(self.device), y.to(self.device)
                self.optimizer.zero_grad()
                y_hat = self.model(X)
                loss = self.hyper_params.loss_fn(y_hat, y)

                # Dynamic regularization
                curr_params = get_all_params_of(self.model, False).to(self.device)
                # penalty = -torch.sum(curr_params * self.prev_grads)
                # penalty += 0.5 * alpha_coef_adpt * torch.sum((curr_params - server_params) ** 2)
                penalty = alpha_coef_adpt * \
                    torch.sum(curr_params * (-server_params + self.prev_grads))
                loss = loss + penalty

                loss.backward()
                self._clip_grads(self.model)
                self.optimizer.step()
                running_loss += loss.item()

            self.scheduler.step()

        # update the previous gradients
        curr_params = get_all_params_of(self.model).to(self.device)
        self.prev_grads += alpha_coef_adpt * (server_params - curr_params)

        running_loss /= (epochs * len(self.train_set))
        self.model.cpu()
        clear_cuda_cache()
        return running_loss


class FedDynServer(Server):
    def __init__(self,
                 model: Module,
                 test_set: FastDataLoader,
                 clients: Iterable[Client],
                 weighted: bool = True,
                 alpha: float = 0.01):
        super().__init__(model=model, test_set=test_set, clients=clients, weighted=weighted)
        self.alpha = alpha
        self.device = FlukeENV().get_device()
        self.cld_mdl = deepcopy(self.model).to(self.device)

    def broadcast_model(self, eligible: Iterable[Client]) -> None:
        self.channel.broadcast(Message((self.model, self.cld_mdl), "model", self), eligible)

    def fit(self, n_rounds: int = 10, eligible_perc: float = 0.1, finalize: bool = True) -> None:

        # Weight computation
        for client in self.clients:
            client._send_weight()

        weights = np.array([self.channel.receive(
            self, client, msg_type="weight").payload for client in self.clients])
        weights = weights / np.sum(weights) * self.n_clients

        for i, client in enumerate(self.clients):
            self.channel.send(Message(weights[i], "weight", self), client)

        for client in self.clients:
            client._receive_weights()

        return super().fit(n_rounds, eligible_perc)

    @torch.no_grad()
    def aggregate(self, eligible: Iterable[Client], client_models: Iterable[Module]) -> None:
        weights = self._get_client_weights(eligible)
        agg_model_sd = aggregate_models(self.model,
                                        client_models,
                                        weights,
                                        eta=self.hyper_params.lr,
                                        inplace=False).state_dict()

        avg_grad = None
        grad_count = 0
        for client in eligible:
            pg = self.channel.receive(self, client, msg_type="grads").payload
            if pg is not None:
                grad_count += 1
                if avg_grad is None:
                    avg_grad = pg.clone().detach()
                else:
                    avg_grad += pg.clone().detach()

        if grad_count > 0:
            avg_grad /= grad_count

        self.model.load_state_dict(agg_model_sd)
        load_all_params(self.device,
                        self.cld_mdl,
                        get_all_params_of(self.model).to(self.device) + avg_grad.to(self.device))


class FedDyn(CentralizedFL):

    def get_client_class(self) -> Client:
        return FedDynClient

    def get_server_class(self) -> Server:
        return FedDynServer



"""Implementation of the FedLD [FedLD24]_ algorithm.

References:
    .. [FedLD24] Shuang Zeng, Pengxin Guo, Shuai Wang, Jianbo Wang, Yuyin Zhou and Liangqiong Qu.
       Tackling Data Heterogeneity in Federated Learning via Loss Decomposition.
       In MICCAI (2024). URL: https://papers.miccai.org/miccai-2024/paper/1348_paper.pdf

"""
import sys
from typing import Any, Iterable

import torch
from torch.nn import CrossEntropyLoss, Module

sys.path.append(".")
sys.path.append("..")

from ..client import Client  # NOQA
from ..data import FastDataLoader  # NOQA
from ..server import Server  # NOQA
from ..utils import OptimizerConfigurator  # NOQA
from . import CentralizedFL  # NOQA

__all__ = [
    "FedLD",
    "FedLDClient",
    "FedLDServer"
]


class MarginalLogLoss(Module):

    def __init__(self,
                 base_loss: Module = CrossEntropyLoss(),
                 lam: float = 0.2):
        super(MarginalLogLoss, self).__init__()
        self.base_loss = base_loss
        self.lam = lam

    def forward(self, y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
        base_loss = self.base_loss(y_pred, y_true)
        softmax_output = torch.softmax(y_pred, dim=1)
        magnitude = torch.norm(softmax_output, p=2)
        marginal_base_loss = base_loss + self.lam * torch.log(1 + magnitude ** 2)
        return marginal_base_loss

    def set_params(self, **kwargs) -> None:
        for key, value in kwargs.items():
            if hasattr(self, key) and value is not None:
                setattr(self, key, value)

    def __str__(self):
        return f"MarginalLogLoss(base_loss={self.base_loss}, lambda={self.lam})"

    def __repr__(self) -> str:
        return str(self)


class FedLDClient(Client):
    def __init__(self,
                 index: int,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: Module,
                 local_epochs: int = 3,
                 fine_tuning_epochs: int = 0,
                 clipping: float = 0,
                 margin_lam: float = 0.2,
                 **kwargs: dict[str, Any]):
        super().__init__(index=index,
                         train_set=train_set,
                         test_set=test_set,
                         optimizer_cfg=optimizer_cfg,
                         loss_fn=loss_fn,
                         local_epochs=local_epochs,
                         fine_tuning_epochs=fine_tuning_epochs,
                         clipping=clipping,
                         **kwargs)
        self.hyper_params.update(margin_lam=margin_lam,
                                 loss_fn=MarginalLogLoss(base_loss=self.hyper_params.loss_fn,
                                                         lam=margin_lam))


def _get_client_grads(client_model: Module, server_model: Module):
    grads = []
    for key in server_model.state_dict().keys():
        grads.append(client_model.state_dict()[key].data.clone().detach().flatten(
        ) - server_model.state_dict()[key].data.clone().detach().flatten())
    return torch.cat(grads)


def _set_client_grads(client_model: Module,
                      server_model: Module,
                      new_grads: torch.Tensor):
    start = 0
    for key in server_model.state_dict().keys():
        dims = client_model.state_dict()[key].shape
        end = start + dims.numel()
        client_model.state_dict()[key].data.copy_(server_model.state_dict()[
            key].data.clone().detach() + new_grads[start:end].reshape(dims).clone())
        start = end
    return client_model


def pcgrad_svd(client_grads: Iterable[Module],
               grad_history: dict,
               k_proportion: float = 0.1):
    """ Projecting conflicting gradients using SVD"""
    client_num = len(client_grads)
    client_grads_ = torch.stack(client_grads)
    grads = []
    grad_len = grad_history['grad_len']
    start = 0
    for key in grad_len.keys():
        g_len = grad_len[key]
        end = start + g_len
        layer_grad_history = grad_history[key]
        client_grads_layer = client_grads_[:, start:end]
        naive_avg_grad = torch.mean(client_grads_layer, dim=0, keepdim=True)
        if layer_grad_history is not None:
            if not torch.all(layer_grad_history == 0):
                hessian = 1 / client_grads_layer.size(0) * client_grads_layer @ client_grads_layer.T
                u, s, e = torch.svd(hessian)
                v = client_grads_layer.T @ e  # (d*m)*(m*m) = (d*m)
                v = v.T  # (m*d)
                k = int(len(s) * k_proportion)
                w = v[:k]
                for j in range(k):
                    num_pos = 0
                    num_neg = 0
                    for i in range(client_grads_layer.size(0)):
                        if torch.dot(client_grads_layer[i], w[j]) >= 0:
                            num_pos += 1
                        else:
                            num_neg += 1
                    if num_pos < num_neg:
                        w[j] = -w[j]
                grad_agg = []
                for i in range(client_num):
                    grad_pc = client_grads_layer[i]
                    grad_revise = torch.zeros_like(grad_pc)
                    for j in range(k):
                        grad_revise_j = torch.dot(grad_pc, w[j])/torch.dot(w[j], w[j]) * w[j]
                        grad_revise_j = grad_revise_j * \
                            s[j]/s.sum() * torch.norm(grad_pc) / torch.norm(grad_revise_j)
                        grad_revise += grad_revise_j
                    grad_agg.append(grad_revise)
                grad_new = torch.mean(torch.stack(grad_agg), dim=0, keepdim=True)
                grad_new = torch.squeeze(grad_new)
                if g_len == 1:
                    grad_new = torch.squeeze(naive_avg_grad)
                    grad_new = torch.unsqueeze(grad_new, 0)
            else:
                client_grads_layer = client_grads_[:, start:end]
                naive_avg_grad = torch.mean(client_grads_layer, dim=0, keepdim=True)
                grad_new = naive_avg_grad
                grad_new = torch.squeeze(grad_new)
                if g_len == 1:
                    grad_new = torch.squeeze(naive_avg_grad)
                    grad_new = torch.unsqueeze(grad_new, 0)
            gamma = 0.99
            grad_history[key] = gamma * grad_history[key] + (1 - gamma) * grad_new
            grads.append(grad_new)
        else:
            grad_new = client_grads_[:, start:end].mean(0)
            grad_history[key] = grad_new
            grads.append(grad_new)
        start = end
    grad_new = torch.cat(grads)

    return grad_new, grad_history


class FedLDServer(Server):

    def __init__(self,
                 model: torch.nn.Module,
                 test_set: FastDataLoader,
                 clients: Iterable[Client],
                 weighted: bool = False,
                 lr: float = 1.0,
                 k_proportion: float = 0.8,
                 **kwargs: dict[str, Any]):
        super().__init__(model=model,
                         test_set=test_set,
                         clients=clients,
                         weighted=weighted,
                         lr=lr,
                         **kwargs)
        self.hyper_params.update(k_proportion=k_proportion)
        self.grad_history = {key: None for key in model.state_dict().keys()}
        self.grad_history['grad_len'] = {key: value.numel()
                                         for key, value in model.state_dict().items()}

    @torch.no_grad()
    def aggregate(self, eligible: Iterable[Client], client_models: Iterable[Module]) -> None:
        client_models = list(client_models)
        local_clients_grads = [_get_client_grads(c_model, self.model) for c_model in client_models]
        grad_new, self.grad_history = pcgrad_svd(local_clients_grads,
                                                 self.grad_history,
                                                 self.hyper_params.k_proportion)

        for c_model in client_models:
            c_model = _set_client_grads(c_model, self.model, grad_new)

        return super().aggregate(eligible, client_models)


class FedLD(CentralizedFL):

    def get_client_class(self):
        return FedLDClient

    def get_server_class(self):
        return FedLDServer
